﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;

namespace PolyLib
{
    /// <summary>
    /// Models a Laurent ComplexPolynomial with coefficients in the field of real numbers
    /// </summary>
    public class ComplexPolynomial : IEquatable<ComplexPolynomial>, IList<ComplexVarTerm>
    {
        #region Constructors
        /// <summary>
        /// Creates an empty ComplexPolynomial
        /// </summary>
        public ComplexPolynomial()
        {
            _terms = new List<ComplexVarTerm>();
            _isReadOnly = false;
        }

        /// <summary>
        /// Creates a new ComplexPolynomial that is equal to a double number
        /// </summary>
        /// <param name="c">The double number</param>
        public ComplexPolynomial(Complex c)
            : this()
        {
            _terms.Add(new ComplexVarTerm(c));
        }

        /// <summary>
        /// Creates a new ComplexPolynomial based on an array of ComplexVarTerm objects
        /// </summary>
        /// <param name="terms">The array of ComplexVarTerm objects</param>
        public ComplexPolynomial(params ComplexVarTerm[] terms)
            : this()
        {
            for (int i = 0; i < terms.Length; i++)
                Add(terms[i].Copy());
        }

        /// <summary>
        /// Creates a new ComplexPolynomial based on a collection of ComplexVarTerm objects
        /// </summary>
        /// <param name="terms">The collection of the ComplexVarTerm objects</param>
        public ComplexPolynomial(IEnumerable<ComplexVarTerm> terms)
            : this()
        {
            foreach (ComplexVarTerm item in terms)
                Add(item);
        }
        #endregion

        #region ICollection
        private bool _isReadOnly;

        /// <summary>
        /// Gets the number of ComplexVarTerm elements contained in the ComplexPolynomial.
        /// </summary>
        public int Count
        {
            get { return _terms.Count; }

        }

        /// <summary>
        /// Gets a value indicating whether the ComplexPolynomial is read-only.
        /// </summary>
        public bool IsReadOnly
        {
            get { return _isReadOnly; }
        }

        /// <summary>
        /// Adds a ComplexVarTerm to the ComplexPolynomial.
        /// </summary>
        /// <param name="item">The ComplexVarTerm object to be added to the ComplexPolynomial.</param>
        public void Add(ComplexVarTerm item)
        {
            if (IsReadOnly) { throw new NotSupportedException("The ComplexPolynomial is read-only."); }
            if (item.Coefficient != 0)
            {
                if (_terms.Contains(item)) //The term is already in the ComplexPolynomial, so it does not need to be added, just add the coefficients
                {
                    int index = _terms.IndexOf(item);
                    if (_terms[index].Coefficient == -item.Coefficient)
                        _terms.RemoveAt(index);
                    else
                        _terms[index].Coefficient += item.Coefficient;
                }
                else
                {
                    _terms.Add(item);
                }
            }
        }

        /// <summary>
        /// Removes all ComplexVarTerm objects from the ComplexPolynomial.
        /// </summary>
        public void Clear()
        {
            if (IsReadOnly) { throw new NotSupportedException("The ComplexPolynomial is read-only."); }
            _terms.Clear();
        }

        /// <summary>
        /// Determines whether the ComplexPolynomial contains a specific ComplexVarTerm (ignores the coefficient).
        /// </summary>
        /// <param name="item">The ComplexVarTerm object to locate in the ComplexPolynomial.</param>
        /// <returns>A boolean value determining whether <paramref name="item"/> is contained in the ComplexPolynomial</returns>
        public bool Contains(ComplexVarTerm item)
        {
            return _terms.Contains(item);
        }

        /// <summary>
        /// Copies the elements of the ComplexPolynomial to an Array, starting at a particular Array index.
        /// </summary>
        /// <param name="array">The one-dimensional Array that is the destination of the elements copied from ComplexPolynomial. The Array must have zero-based indexing.</param>
        /// <param name="arrayIndex">The zero-based index in array at which copying begins.</param>
        public void CopyTo(ComplexVarTerm[] array, int arrayIndex)
        {
            if (array.Length == 0) { throw new ArgumentNullException("array", "array is null"); }
            if (arrayIndex < 0) { throw new ArgumentOutOfRangeException("arrayIndex", arrayIndex, "arrayIndex is less than 0"); }
            if (array.Length - arrayIndex < _terms.Count) { throw new ArgumentException("The number of elements in the source ComplexPolynomial is greater than the available space from arrayIndex to the end of the destination array."); }

            for (int i = 0; i < _terms.Count; i++)
                array[arrayIndex + i] = _terms[i].Copy();
        }

        /// <summary>
        /// Removes a specific ComplexVarTerm from the ComplexPolynomial.
        /// </summary>
        /// <param name="item">The object to remove from the ComplexPolynomial.</param>
        /// <returns>true if item was successfully removed from the ComplexPolynomial; otherwise, false. This method also returns false if item is not found in the original ComplexPolynomial.</returns>
        public bool Remove(ComplexVarTerm item)
        {
            if (IsReadOnly) { throw new NotSupportedException("The ComplexPolynomial is read-only."); }
            bool rem = _terms.Remove(item);
            return rem;
        }
        #endregion

        #region IEnumerable
        /// <summary>
        /// Gets an Enumerator for the ComplexPolynomial object
        /// </summary>
        /// <returns>An IEnumerator<ComplexVarTerm> object</ComplexVarTerm></returns>
        public IEnumerator<ComplexVarTerm> GetEnumerator()
        {
            return _terms.GetEnumerator();
        }

        /// <summary>
        /// Gets an Enumerator for the ComplexPolynomial object
        /// </summary>
        /// <returns>An IEnumerator object</ComplexVarTerm></returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return (IEnumerator)GetEnumerator();
        }
        #endregion

        #region IList
        private List<ComplexVarTerm> _terms;

        /// <summary>
        /// Gets or sets the ComplexVarTerm at the specified index.
        /// </summary>
        /// <param name="index">The zero-based ComplexVarTerm of the element to get or set.</param>
        /// <returns>The ComplexVarTerm at the specified index.</returns>
        public ComplexVarTerm this[int index]
        {
            get { return _terms[index]; }
            set
            {
                if (value.Coefficient != Complex.Zero)
                {
                    _terms[index] = value;
                }
                else
                {
                    if (index > -1 && index < _terms.Count) { _terms.RemoveAt(index); }
                }
            }
        }

        /// <summary>
        /// Determines the index of a specific item in the ComplexPolynomial.
        /// </summary>
        /// <param name="t">The ComplexVarTerm object to locate in the ComplexPolynomial.</param>
        /// <returns>The index of ComplexVarTerm if found in the ComplexPolynomial; otherwise, -1.</returns>
        public int IndexOf(ComplexVarTerm item)
        {
            return _terms.IndexOf(item);
        }

        /// <summary>
        /// Adds a ComplexVarTerm to the Polynomia
        /// </summary>
        /// <param name="index">The zero-based index at which item should be inserted</param>
        /// <param name="item">The object to insert into the ComplexPolynomial</param>
        public void Insert(int index, ComplexVarTerm item)
        {
            _terms.Insert(index, item);
        }

        /// <summary>
        /// Removes the ComplexVarTerm at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the ComplexVarTerm to remove.</param>
        public void RemoveAt(int index)
        {
            if (IsReadOnly) { throw new NotSupportedException("The ComplexPolynomial is read-only."); }
            _terms.RemoveAt(index);
        }
        #endregion

        #region IEquatable
        /// <summary>
        /// Checks if the current ComplexPolynomial is equal to a given ComplexPolynomial
        /// </summary>
        /// <param name="p">A ComplexPolynomial</param>
        /// <returns>A boolean value determining whether the current ComplexPolynomial is equal to <paramref name="p"/></returns>
        public Boolean Equals(ComplexPolynomial p)
        {
            return (this.ToString() == p.ToString());
        }
        #endregion

        #region Operators
        /// <summary>
        /// Multiplies a ComplexPolynomial with a ComplexPolynomial   
        /// </summary>
        /// <param name="p1">The first ComplexPolynomial</param>
        /// <param name="p2">The second ComplexPolynomial</param>
        /// <returns>The product of the two objects</returns>
        public static ComplexPolynomial operator *(ComplexPolynomial p1, ComplexPolynomial p2)
        {
            if (p1._terms.Count == 0) { return p2.Copy(); }
            if (p2._terms.Count == 0) { return p1.Copy(); }
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm m1 in p1._terms)
                foreach (ComplexVarTerm m2 in p2._terms)
                    temp.Add(m1 * m2);
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Multiplies a ComplexPolynomial with a ComplexVarTerm
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="t">The ComplexVarTerm</param>
        /// <returns>The product of the two objects</returns>
        public static ComplexPolynomial operator *(ComplexPolynomial p, ComplexVarTerm t)
        {
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm m1 in p._terms)
                temp.Add(m1 * t);
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Multiplies a ComplexPolynomial with a ComplexVarTerm
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="t">The ComplexVarTerm</param>
        /// <returns>The product of the two objects</returns>
        public static ComplexPolynomial operator *(ComplexVarTerm t, ComplexPolynomial p)
        {
            return p * t;
        }

        /// <summary>
        /// Multiplies a ComplexPolynomial with a double number
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The product of the two objects</returns>
        public static ComplexPolynomial operator *(ComplexPolynomial p, Complex c)
        {
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm m1 in p._terms)
                temp.Add(m1 * c);
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Multiplies a ComplexPolynomial with a double number
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The product of the two objects</returns>
        public static ComplexPolynomial operator *(Complex c, ComplexPolynomial p)
        {
            return p * c;
        }

        /// <summary>
        /// Adds a ComplexPolynomial with a ComplexPolynomial
        /// </summary>
        /// <param name="p1">The first ComplexPolynomial</param>
        /// <param name="p2">The second ComplexPolynomial</param>
        /// <returns>The sum of the two objects</returns>
        public static ComplexPolynomial operator +(ComplexPolynomial p1, ComplexPolynomial p2)
        {
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm m in p1._terms) //Adds the terms of p1
                temp.Add(m.Copy());
            foreach (ComplexVarTerm m in p2._terms) //Adds the terms of p2
                temp.Add(m.Copy());
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Adds a ComplexPolynomial with a ComplexVarTerm
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="t">The ComplexVarTerm</param>
        /// <returns>The sum of the two objects</returns>
        public static ComplexPolynomial operator +(ComplexPolynomial p, ComplexVarTerm t)
        {
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm m in p._terms)
                temp.Add(m.Copy());
            temp.Add(t.Copy());
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Adds a ComplexPolynomial with a ComplexVarTerm
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="t">The ComplexVarTerm</param>
        /// <returns>The sum of the two objects</returns>
        public static ComplexPolynomial operator +(ComplexVarTerm t, ComplexPolynomial p)
        {
            return p + t;
        }

        /// <summary>
        /// Adds a ComplexPolynomial with a double number
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The sum of the two objects</returns>
        public static ComplexPolynomial operator +(ComplexPolynomial p, Complex c)
        {
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm m in p._terms)
                temp.Add(m.Copy());
            temp.Add(new ComplexVarTerm(c));
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Adds a ComplexPolynomial with a double number
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The sum of the two objects</returns>
        public static ComplexPolynomial operator +(Complex c, ComplexPolynomial p)
        {
            return p + c;
        }

        /// <summary>
        /// Calculates the opposite ComplexPolynomial
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <returns>The opposite ComplexPolynomial</returns>
        public static ComplexPolynomial operator -(ComplexPolynomial p)
        {
            List<ComplexVarTerm> temp = new List<ComplexVarTerm>();
            foreach (ComplexVarTerm t in p._terms) //Calculates the opposite terms
                temp.Add(-t.Copy());
            return new ComplexPolynomial(temp);
        }

        /// <summary>
        /// Substracts a ComplexPolynomial with a ComplexPolynomial
        /// </summary>
        /// <param name="p1">The first ComplexPolynomial</param>
        /// <param name="p2">The second ComplexPolynomial</param>
        /// <returns>The difference of the two objects</returns>
        public static ComplexPolynomial operator -(ComplexPolynomial p1, ComplexPolynomial p2)
        {
            return p1 + (-p2);
        }

        /// <summary>
        /// Substracts a ComplexPolynomial with a ComplexVarTerm
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="t">The ComplexVarTerm</param>
        /// <returns>The difference of the two objects</returns>
        public static ComplexPolynomial operator -(ComplexPolynomial p, ComplexVarTerm t)
        {
            return p + (-t);
        }

        /// <summary>
        /// Substracts a ComplexPolynomial with a ComplexVarTerm
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="t">The ComplexVarTerm</param>
        /// <returns>The difference of the two objects</returns>
        public static ComplexPolynomial operator -(ComplexVarTerm t, ComplexPolynomial p)
        {
            return -p + t;
        }

        /// <summary>
        /// Substracts a ComplexPolynomial with a double number
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The difference of the two objects</returns>
        public static ComplexPolynomial operator -(ComplexPolynomial p1, Complex c)
        {
            return p1 + (-c);
        }

        /// <summary>
        /// Substracts a ComplexPolynomial with a double number
        /// </summary>
        /// <param name="p">The ComplexPolynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The difference of the two objects</returns>
        public static ComplexPolynomial operator -(Complex c, ComplexPolynomial p)
        {
            return p - c;
        }
        #endregion

        #region Other Methods
        /// <summary>
        /// Returns the ComplexPolynomial in the form of a string
        /// </summary>
        /// <returns>A string</returns>
        public override string ToString()
        {
            Sort();
            StringBuilder str = new StringBuilder();

            //bool first = true;
            str.Append(_terms[0].ToString());
            for (int i = 1; i < _terms.Count; i++)
            {
                //if (Math.Sign(_terms[i].Coefficient) == 1 && !first) { str.Append(" + "); }
                //first = false;
                str.Append(" + " + _terms[i].ToString());
            }
            
            return str.ToString();
        }

        /// <summary>
        /// Creates a deep copy of the ComplexPolynomial
        /// </summary>
        /// <returns>A copy of the ComplexPolynomial</returns>
        public ComplexPolynomial Copy()
        {
            ComplexPolynomial copy = new ComplexPolynomial();

            foreach (ComplexVarTerm term in this._terms)
                copy._terms.Add(term.Copy());
            //copy._terms.Sort();
            return copy;
        }

        /// <summary>
        /// Sorts the terms of the current ComplexPolynomial
        /// </summary>
        public void Sort()
        {
            _terms.Sort();
        }

        /// <summary>
        /// Sets the read-only property of the ComplexPolynomial
        /// </summary>
        /// <param name="readOnly">A boolean value determining whether the ComplexPolynomial will be read-only</param>
        public void SetReadOnly(bool readOnly)
        {
            _isReadOnly = readOnly;
        }
        #endregion
    }
}
