﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;

namespace PolyLib
{
    /// <summary>
    /// Models a single term of the polynomial in the field of real numbers
    /// </summary>
    public class ComplexVarTerm : IEquatable<ComplexVarTerm>, IComparable<ComplexVarTerm>
    {
        public SortedDictionary<char, int> Variables { get; set; } //Saves the variables and their exponents
        public Complex Coefficient { get; set; } //The coefficient of the term

        #region Constructors
        /// <summary>
        /// Creates a new ComplexVarTerm which equals to the given <paramref name="coeff"/>
        /// </summary>
        /// <param name="coeff">The coefficient of the ComplexVarTerm</param>
        public ComplexVarTerm(Complex coeff)
        {
            Variables = new SortedDictionary<char, int>();
            Coefficient = coeff;
        }

        /// <summary>
        /// Creates a new ComplexVarTerm from a given SortedDictionary and a given coefficient
        /// </summary>
        /// <param name="var">A SortedDictionary object containing the variables and their exponents</param>
        /// <param name="coeff">The coefficient of the ComplexVarTerm</param>
        public ComplexVarTerm(SortedDictionary<char, int> var, Complex coeff) 
            : this(coeff)
        {
            if (coeff != 0)
                //Deep copy of dictionary
                foreach (char k in var.Keys)
                    if (var[k] != 0)
                        Variables.Add(k, var[k]);
            else
                Variables = new SortedDictionary<char, int>();
        }

        /// <summary>
        /// Creates a new ComplexVarTerm from an array of tuples and a given coefficient
        /// </summary>
        /// <param name="coeff">The coefficient of the ComplexVarTerm</param>
        /// <param name="exp">An array of tuples containing the variables and their exponents</param>
        public ComplexVarTerm(Complex coeff, params Tuple<char, int>[] exp) 
            : this(coeff)
        {
            if (coeff != 0)
                for (int i = 0; i < exp.Length; i++)
                    if (exp[i].Item2 != 0)
                        Variables.Add(exp[i].Item1, exp[i].Item2);
            else
                Variables = new SortedDictionary<char, int>();
        }
       
        /// <summary>
        /// Creates a new ComplexVarTerm from a given coefficient, a given character (variable) and a given exponent
        /// </summary>
        /// <param name="coeff">The coefficient</param>
        /// <param name="c">The variable</param>
        /// <param name="exp">The exponent</param>
        public ComplexVarTerm(Complex coeff, char c, int exp)
            : this(coeff, new Tuple<char, int>(c, exp)) { }
        #endregion

        #region Interfaces
        /// <summary>
        /// Checks if a given ComplexVarTerm is equal to the current ComplexVarTerm
        /// </summary>
        /// <param name="t">The ComplexVarTerm to be compared to the current ComplexVarTerm</param>
        /// <returns>A boolean value determining whether the given ComplexVarTerm equals the current ComplexVarTerm</returns>
        public Boolean Equals(ComplexVarTerm t)
        {
            return CompareTo(t) == 0;
        }

        /// <summary>
        /// Compares the current ComplexVarTerm with an other one, based on lexicographical sorting but ignores coefficients
        /// </summary>
        /// <param name="t">ComplexVarTerm</param>
        /// <returns>An integer value that indicates the relative order of the objects being compared.</returns>
        public int CompareTo(ComplexVarTerm t)
        {
            int a = ToStringOnlyVar().CompareTo(t.ToStringOnlyVar());
            if (a == 0)
                return ToStringOnlyExp().CompareTo(t.ToStringOnlyExp());
            else
                return a;
            
        }
        #endregion

        #region Operators
        /// <summary>
        /// Multiplies two ComplexVarTerm objects
        /// </summary>
        /// <param name="t1">The first ComplexVarTerm object</param>
        /// <param name="t2">The second ComplexVarTerm object</param>
        /// <returns>The product of the two ComplexVarTerm objects</returns>
        public static ComplexVarTerm operator *(ComplexVarTerm t1, ComplexVarTerm t2)
        {
            if (t1.Variables.Count == 0) //t1 = Complex number
                return new ComplexVarTerm(t2.Variables, t1.Coefficient * t2.Coefficient);
            else if (t2.Variables.Count == 0) //t2 = Complex number
                return new ComplexVarTerm(t1.Variables, t1.Coefficient * t2.Coefficient);
            else
            {
                ComplexVarTerm result = new ComplexVarTerm(t1.Variables, t1.Coefficient * t2.Coefficient); //Creates a new ComplexVarTerm based on t1
                //Copies the variables and the exponents of t2 into result
                foreach (char c in t2.Variables.Keys)
                {
                    if (result.Variables.ContainsKey(c))
                        result.Variables[c] += t2.Variables[c];
                    else
                        result.Variables.Add(c, t2.Variables[c]);
                }
                return result;
            }
        }

        /// <summary>
        /// Multiplies a ComplexVarTerm with a Complex number
        /// </summary>
        /// <param name="t">The ComplexVarTerm object</param>
        /// <param name="c">The Complex number</param>
        /// <returns>The prodcut of the ComplexVarTerm with the number</returns>
        public static ComplexVarTerm operator *(ComplexVarTerm t, Complex c)
        {
            return new ComplexVarTerm(t.Variables, t.Coefficient * c);
        }

        /// <summary>
        /// Multiplies a ComplexVarTerm with a Complex number
        /// </summary>
        /// <param name="t">The ComplexVarTerm object</param>
        /// <param name="c">The Complex number</param>
        /// <returns>The prodcut of the ComplexVarTerm with the number</returns>
        public static ComplexVarTerm operator *(Complex c, ComplexVarTerm t)
        {
            return t * c;
        }
        
        /// <summary>
        /// Calculates the opposite of a ComplexVarTerm
        /// </summary>
        /// <param name="t">The ComplexVarTerm object</param>
        /// <returns>The opposite ComplexVarTerm object</returns>
        public static ComplexVarTerm operator -(ComplexVarTerm t)
        {
            return new ComplexVarTerm(t.Variables, -t.Coefficient);
        }
        #endregion

        #region Other Methods
        /// <summary>
        /// Adds a number to the coefficient
        /// </summary>
        /// <param name="c">The number to be added</param>
        public void AddToCoeff(Complex c)
        {
            Coefficient += c;
        }

        /// <summary>
        /// Returns the ComplexVarTerm in the form of a string
        /// </summary>
        /// <returns>A string</returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();
            str.Append(Coefficient);
            //if (Coefficient < 0)
            //{
            //    if (Coefficient == -1 && Variables.Count == 0)
            //    {
            //        str.Append(" - 1");
            //        return str.ToString();
            //    }
            //    else if (Coefficient == -1 && Variables.Count > 0)
            //    {
            //        str.Append(" -");
            //    }
            //    else
            //        str.Append(" - " + -Coefficient);
            //}
            //else if (Coefficient > 0)
            //{
            //    if (!(Coefficient == 1 && Variables.Count != 0))
            //        str.Append(Coefficient);
            //}
            
                
            foreach (KeyValuePair<char, int> k in Variables)
            {
                if (str.Length > 0) { str.Append(" "); }
                str.Append(k.Key.ToString());
                if (k.Value != 1) { str.Append("^" + k.Value); }
            }
            return str.ToString();
        }

        /// <summary>
        /// Returns the ComplexVarTerm in the form of a string (ignores the coefficients)
        /// </summary>
        /// <returns>A string</returns>
        public string ToStringOnlyVar()
        {
            StringBuilder str = new StringBuilder();
            foreach (KeyValuePair<char, int> k in Variables)
            {
                str.Append(k.Key.ToString());
            }
            return str.ToString();
        }

        /// <summary>
        /// Returns the ComplexVarTerm in the form of a string (ignores the coefficients)
        /// </summary>
        /// <returns>A string</returns>
        public string ToStringOnlyExp()
        {
            StringBuilder str = new StringBuilder();
            foreach (KeyValuePair<char, int> k in Variables)
            {
                str.Append(k.Value);
            }
            return str.ToString();
        }

        /// <summary>
        /// Creates a deep copy
        /// </summary>
        /// <returns>A new ComplexVarTerm</returns>
        public ComplexVarTerm Copy()
        {
            ComplexVarTerm copy = new ComplexVarTerm(this.Coefficient);

            foreach (char k in this.Variables.Keys)
                copy.Variables.Add(k, this.Variables[k]);

            return copy;
        }
        #endregion
    }
}
