﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;

namespace PolyLib
{
    /// <summary>
    /// Models a single term of the polynomial in the field of real numbers
    /// </summary>
    public class MultiVarTerm : IEquatable<MultiVarTerm>, IComparable<MultiVarTerm>
    {
        public SortedDictionary<char, int> Variables; //Saves the variables and their exponents
        public double Coefficient; //The coefficient of the term

        #region Constructors

        /// <summary>
        /// Creates a new MultiVarTerm which equals to the given <paramref name="coeff"/>
        /// </summary>
        /// <param name="coeff">The coefficient of the MultiVarTerm</param>
        public MultiVarTerm(double coeff = 1) 
        {
            Variables = new SortedDictionary<char, int>();
            Coefficient = coeff;
        }

        /// <summary>
        /// Creates a new MultiVarTerm from a given SortedDictionary and a given coefficient
        /// </summary>
        /// <param name="var">A SortedDictionary object containing the variables and their exponents</param>
        /// <param name="coeff">The coefficient of the MultiVarTerm</param>
        public MultiVarTerm(SortedDictionary<char, int> var, double coeff = 1) 
            : this(coeff)
        {
            if (coeff != 0)
                //Deep copy of dictionary
                foreach (char k in var.Keys)
                    if (var[k] != 0)
                        Variables.Add(k, var[k]);
            else
                Variables = new SortedDictionary<char, int>();
        }

        /// <summary>
        /// Creates a new MultiVarTerm from an array of tuples and a given coefficient
        /// </summary>
        /// <param name="coeff">The coefficient of the MultiVarTerm</param>
        /// <param name="exp">An array of tuples containing the variables and their exponents</param>
        public MultiVarTerm(double coeff = 1, params Tuple<char, int>[] exp) 
            : this(coeff)
        {
            if (coeff != 0)
                for (int i = 0; i < exp.Length; i++)
                    if (exp[i].Item2 != 0)
                        Variables.Add(exp[i].Item1, exp[i].Item2);
            else
                Variables = new SortedDictionary<char, int>();
        }
       
        /// <summary>
        /// Creates a new MultiVarTerm from a given coefficient, a given character (variable) and a given exponent
        /// </summary>
        /// <param name="coeff">The coefficient</param>
        /// <param name="c">The variable</param>
        /// <param name="exp">The exponent</param>
        public MultiVarTerm(double coeff, char c, int exp)
            : this(coeff, new Tuple<char, int>(c, exp)) { }
        #endregion

        #region Interfaces
        /// <summary>
        /// Checks if a given MultiVarTerm is equal to the current MultiVarTerm
        /// </summary>
        /// <param name="t">The MultiVarTerm to be compared to the current MultiVarTerm</param>
        /// <returns>A boolean value determining whether the given MultiVarTerm equals the current MultiVarTerm</returns>
        public Boolean Equals(MultiVarTerm t)
        {
            return CompareTo(t) == 0;
        }

        /// <summary>
        /// Compares the current MultiVarTerm with an other one, based on lexicographical sorting but ignores coefficients
        /// </summary>
        /// <param name="t">MultiVarTerm</param>
        /// <returns>An integer value that indicates the relative order of the objects being compared.</returns>
        public int CompareTo(MultiVarTerm t)
        {
            int a = ToStringOnlyVar().CompareTo(t.ToStringOnlyVar());
            if (a == 0)
                return ToStringOnlyExp().CompareTo(t.ToStringOnlyExp());
            else
                return a;
            
        }
        #endregion

        #region Operators
        /// <summary>
        /// Multiplies two MultiVarTerm objects
        /// </summary>
        /// <param name="t1">The first MultiVarTerm object</param>
        /// <param name="t2">The second MultiVarTerm object</param>
        /// <returns>The product of the two MultiVarTerm objects</returns>
        public static MultiVarTerm operator *(MultiVarTerm t1, MultiVarTerm t2)
        {
            if (t1.Variables.Count == 0) //t1 = double number
                return new MultiVarTerm(t2.Variables, t1.Coefficient * t2.Coefficient);
            else if (t2.Variables.Count == 0) //t2 = double number
                return new MultiVarTerm(t1.Variables, t1.Coefficient * t2.Coefficient);
            else
            {
                MultiVarTerm result = new MultiVarTerm(t1.Variables, t1.Coefficient * t2.Coefficient); //Creates a new MultiVarTerm based on t1
                //Copies the variables and the exponents of t2 into result
                foreach (char c in t2.Variables.Keys)
                {
                    if (result.Variables.ContainsKey(c))
                        result.Variables[c] += t2.Variables[c];
                    else
                        result.Variables.Add(c, t2.Variables[c]);
                }
                foreach (char c in t2.Variables.Keys)
                {
                    if (result.Variables[c] == 0)
                        result.Variables.Remove(c);
                }
                return result;
            }
        }

        /// <summary>
        /// Multiplies a MultiVarTerm with a double number
        /// </summary>
        /// <param name="t">The MultiVarTerm object</param>
        /// <param name="c">The double number</param>
        /// <returns>The prodcut of the MultiVarTerm with the number</returns>
        public static MultiVarTerm operator *(MultiVarTerm t, double c)
        {
            return new MultiVarTerm(t.Variables, t.Coefficient * c);
        }

        /// <summary>
        /// Multiplies a MultiVarTerm with a double number
        /// </summary>
        /// <param name="t">The MultiVarTerm object</param>
        /// <param name="c">The double number</param>
        /// <returns>The prodcut of the MultiVarTerm with the number</returns>
        public static MultiVarTerm operator *(double c, MultiVarTerm t)
        {
            return t * c;
        }
        
        /// <summary>
        /// Calculates the opposite of a MultiVarTerm
        /// </summary>
        /// <param name="t">The MultiVarTerm object</param>
        /// <returns>The opposite MultiVarTerm object</returns>
        public static MultiVarTerm operator -(MultiVarTerm t)
        {
            return new MultiVarTerm(t.Variables, -t.Coefficient);
        }
        #endregion

        #region Other Methods
        /// <summary>
        /// Adds a number to the coefficient
        /// </summary>
        /// <param name="c">The number to be added</param>
        public void AddToCoeff(double c)
        {
            Coefficient += c;
        }

        /// <summary>
        /// Returns the MultiVarTerm in the form of a string
        /// </summary>
        /// <returns>A string</returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();

            if (Coefficient < 0)
            {
                if (Coefficient == -1 && Variables.Count == 0)
                {
                    str.Append(" - 1");
                    return str.ToString();
                }
                else if (Coefficient == -1 && Variables.Count > 0)
                {
                    str.Append(" -");
                }
                else
                    str.Append(" - " + -Coefficient);
            }
            else if (Coefficient > 0)
            {
                if (!(Coefficient == 1 && Variables.Count != 0))
                    str.Append(Coefficient);
            }
            
                
            foreach (KeyValuePair<char, int> k in Variables)
            {
                if (str.Length > 0) { str.Append(" "); }
                str.Append(k.Key.ToString());
                if (k.Value != 1) { str.Append("^" + k.Value); }
            }
            return str.ToString();
        }

        /// <summary>
        /// Returns the MultiVarTerm in the form of a string
        /// </summary>
        /// <returns>A string</returns>
        public string ToStringMaple()
        {
            StringBuilder str = new StringBuilder();

            if (Coefficient < 0)
            {
                if (Coefficient == -1 && Variables.Count == 0)
                {
                    str.Append(" - 1");
                    return str.ToString();
                }
                else if (Coefficient == -1 && Variables.Count > 0)
                {
                    str.Append(" -");
                }
                else
                    str.Append(" - " + -Coefficient);
            }
            else if (Coefficient > 0)
            {
                if (!(Coefficient == 1 && Variables.Count != 0))
                    str.Append(Coefficient);
            }


            foreach (KeyValuePair<char, int> k in Variables)
            {
                if (str.Length >  0 && str[str.Length - 1] != '-') { str.Append("*"); }
                //if (str.Length > 0) { str.Append(""); }
                if (k.Key > 64 && k.Key < 91)
                    str.Append("x[" + (k.Key - 64).ToString() + "]");
                else
                    str.Append(k.Key.ToString());
                if (k.Value != 1) { str.Append("^(" + k.Value + ")"); }
            }
            return str.ToString();
        }

        /// <summary>
        /// Returns the MultiVarTerm in the form of a string (ignores the coefficients)
        /// </summary>
        /// <returns>A string</returns>
        public string ToStringOnlyVar()
        {
            StringBuilder str = new StringBuilder();
            foreach (KeyValuePair<char, int> k in Variables)
            {
                str.Append(k.Key.ToString());
            }
            return str.ToString();
        }

        /// <summary>
        /// Returns the MultiVarTerm in the form of a string (ignores the coefficients)
        /// </summary>
        /// <returns>A string</returns>
        public string ToStringOnlyExp()
        {
            StringBuilder str = new StringBuilder();
            foreach (KeyValuePair<char, int> k in Variables)
            {
                str.Append(k.Value);
            }
            return str.ToString();
        }

        /// <summary>
        /// Creates a deep copy
        /// </summary>
        /// <returns>A new MultiVarTerm</returns>
        public MultiVarTerm Copy()
        {
            MultiVarTerm copy = new MultiVarTerm(this.Coefficient);

            foreach (char k in this.Variables.Keys)
                copy.Variables.Add(k, this.Variables[k]);

            return copy;
        }
        #endregion
    }
}
