﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PolyLib
{
    public static class Time
    {
        public static double Milliseconds = 0;
    }

    /// <summary>
    /// Models a Laurent polynomial with coefficients in the field of real numbers
    /// </summary>
    public class Polynomial :IEquatable<Polynomial>, IList<MultiVarTerm>
    {
        #region Constructors
        /// <summary>
        /// Creates an empty Polynomial
        /// </summary>
        public Polynomial()
        {
            _terms = new List<MultiVarTerm>();
            _isReadOnly = false;
        }

        /// <summary>
        /// Creates a new Polynomial that is equal to a double number
        /// </summary>
        /// <param name="c">The double number</param>
        public Polynomial(double c) : this()
        {
            _terms.Add(new MultiVarTerm(c));
        }

        /// <summary>
        /// Creates a new Polynomial based on an array of MultiVarTerm objects
        /// </summary>
        /// <param name="terms">The array of MultiVarTerm objects</param>
        public Polynomial(params MultiVarTerm[] terms) : this()
        {
            for (int i = 0; i < terms.Length; i++)
                Add(terms[i].Copy());
        }

        /// <summary>
        /// Creates a new Polynomial based on a collection of MultiVarTerm objects
        /// </summary>
        /// <param name="terms">The collection of the MultiVarTerm objects</param>
        public Polynomial(IEnumerable<MultiVarTerm> terms) : this()
        {
            foreach (MultiVarTerm item in terms)
                Add(item);
        }
        #endregion

        #region ICollection
        private bool _isReadOnly;

        /// <summary>
        /// Gets the number of MultiVarTerm elements contained in the Polynomial.
        /// </summary>
        public int Count
        {
            get { return _terms.Count; }

        }

        /// <summary>
        /// Gets a value indicating whether the Polynomial is read-only.
        /// </summary>
        public bool IsReadOnly
        {
            get { return _isReadOnly; }
        }
        
        /// <summary>
        /// Adds a MultiVarTerm to the Polynomial.
        /// </summary>
        /// <param name="item">The MultiVarTerm object to be added to the Polynomial.</param>
        public void Add(MultiVarTerm item)
        {
            if (IsReadOnly) { throw new NotSupportedException("The Polynomial is read-only."); }
            if (item.Coefficient != 0)
            {
                if (_terms.Contains(item)) //The term is already in the Polynomial, so it does not need to be added, just add the coefficients
                {
                    int index = _terms.IndexOf(item);
                    if (_terms[index].Coefficient == -item.Coefficient)
                        _terms.RemoveAt(index);
                    else
                        _terms[index].Coefficient += item.Coefficient;
                }
                else
                {
                    _terms.Add(item);
                }
            }
        }

        /// <summary>
        /// Removes all MultiVarTerm objects from the Polynomial.
        /// </summary>
        public void Clear()
        {
            if (IsReadOnly) { throw new NotSupportedException("The Polynomial is read-only."); }
            _terms.Clear();
        }

        /// <summary>
        /// Determines whether the Polynomial contains a specific MultiVarTerm (ignores the coefficient).
        /// </summary>
        /// <param name="item">The MultiVarTerm object to locate in the Polynomial.</param>
        /// <returns>A boolean value determining whether <paramref name="item"/> is contained in the Polynomial</returns>
        public bool Contains(MultiVarTerm item)
        {
            return _terms.Contains(item);
        }

        /// <summary>
        /// Copies the elements of the Polynomial to an Array, starting at a particular Array index.
        /// </summary>
        /// <param name="array">The one-dimensional Array that is the destination of the elements copied from Polynomial. The Array must have zero-based indexing.</param>
        /// <param name="arrayIndex">The zero-based index in array at which copying begins.</param>
        public void CopyTo(MultiVarTerm[] array, int arrayIndex)
        {
            if (array.Length == 0) { throw new ArgumentNullException("array", "array is null"); }
            if (arrayIndex < 0) { throw new ArgumentOutOfRangeException("arrayIndex", arrayIndex, "arrayIndex is less than 0"); }
            if (array.Length - arrayIndex < _terms.Count) { throw new ArgumentException("The number of elements in the source Polynomial is greater than the available space from arrayIndex to the end of the destination array."); }

            for (int i = 0; i < _terms.Count; i++)
                array[arrayIndex + i] = _terms[i].Copy();
        }     
        
        /// <summary>
        /// Removes a specific MultiVarTerm from the Polynomial.
        /// </summary>
        /// <param name="item">The object to remove from the Polynomial.</param>
        /// <returns>true if item was successfully removed from the Polynomial; otherwise, false. This method also returns false if item is not found in the original Polynomial.</returns>
        public bool Remove(MultiVarTerm item)
        {
            if (IsReadOnly) { throw new NotSupportedException("The Polynomial is read-only."); }
            bool rem = _terms.Remove(item);
            return rem;
        }
        #endregion

        #region IEnumerable
        /// <summary>
        /// Gets an Enumerator for the Polynomial object
        /// </summary>
        /// <returns>An IEnumerator<MultiVarTerm> object</MultiVarTerm></returns>
        public IEnumerator<MultiVarTerm> GetEnumerator()
        {
            return _terms.GetEnumerator();
        }

        /// <summary>
        /// Gets an Enumerator for the Polynomial object
        /// </summary>
        /// <returns>An IEnumerator object</MultiVarTerm></returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return (IEnumerator) GetEnumerator();
        }
        #endregion

        #region IList
        private List<MultiVarTerm> _terms;

        /// <summary>
        /// Gets or sets the MultiVarTerm at the specified index.
        /// </summary>
        /// <param name="index">The zero-based MultiVarTerm of the element to get or set.</param>
        /// <returns>The MultiVarTerm at the specified index.</returns>
        public MultiVarTerm this[int index]
        {
            get { return _terms[index]; }
            set { _terms[index] = value; }
        }

        /// <summary>
        /// Determines the index of a specific item in the Polynomial.
        /// </summary>
        /// <param name="t">The MultiVarTerm object to locate in the Polynomial.</param>
        /// <returns>The index of MultiVarTerm if found in the Polynomial; otherwise, -1.</returns>
        public int IndexOf(MultiVarTerm item)
        {
            return _terms.IndexOf(item);
        }

        /// <summary>
        /// Adds a MultiVarTerm to the Polynomia
        /// </summary>
        /// <param name="index">The zero-based index at which item should be inserted</param>
        /// <param name="item">The object to insert into the Polynomial</param>
        public void Insert(int index, MultiVarTerm item)
        {
            _terms.Insert(index, item);
        }

        /// <summary>
        /// Removes the MultiVarTerm at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the MultiVarTerm to remove.</param>
        public void RemoveAt(int index)
        {
            if (IsReadOnly) { throw new NotSupportedException("The Polynomial is read-only."); }
            _terms.RemoveAt(index);
        }
        #endregion

        #region IEquatable
        /// <summary>
        /// Checks if the current Polynomial is equal to a given Polynomial
        /// </summary>
        /// <param name="p">A Polynomial</param>
        /// <returns>A boolean value determining whether the current Polynomial is equal to <paramref name="p"/></returns>
        public Boolean Equals(Polynomial p)
        {
            return (this.ToString() == p.ToString());
        }
        #endregion

        #region Operators
        /// <summary>
        /// Multiplies a Polynomial with a Polynomial   
        /// </summary>
        /// <param name="p1">The first Polynomial</param>
        /// <param name="p2">The second Polynomial</param>
        /// <returns>The product of the two objects</returns>
        public static Polynomial operator *(Polynomial p1, Polynomial p2)
        {
            //DateTime dt1 = DateTime.Now;
            if (p1._terms.Count == 1 && p1[0].Variables.Count == 0)
            {
                if (p1[0].Coefficient == 1) { return p2; }
                else { return p1[0].Coefficient * p2; }
            }
            if (p2._terms.Count == 1 && p2[0].Variables.Count == 0)
            {
                if (p2[0].Coefficient == 1) { return p1; }
                else { return p2[0].Coefficient * p1; }
            }
            if (p1._terms.Count == 0) { return p2.Copy(); }
            if (p2._terms.Count == 0) { return p1.Copy(); }
            Polynomial temp =  new Polynomial();
            foreach (MultiVarTerm m1 in p1._terms)
                foreach (MultiVarTerm m2 in p2._terms)
                    temp.Add(m1 * m2);

            //DateTime dt2 = DateTime.Now;
            //Time.Milliseconds += (dt2 - dt1).TotalMilliseconds;
            return temp;
        }

        /// <summary>
        /// Multiplies a Polynomial with a MultiVarTerm
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="t">The MultiVarTerm</param>
        /// <returns>The product of the two objects</returns>
        public static Polynomial operator *(Polynomial p, MultiVarTerm t)
        {
            Polynomial temp = new Polynomial();
            foreach (MultiVarTerm m1 in p._terms)
                    temp.Add(m1 * t);
            return temp;
        }

        /// <summary>
        /// Multiplies a Polynomial with a MultiVarTerm
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="t">The MultiVarTerm</param>
        /// <returns>The product of the two objects</returns>
        public static Polynomial operator *(MultiVarTerm t, Polynomial p)
        {
            return p * t;
        }

        /// <summary>
        /// Multiplies a Polynomial with a double number
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The product of the two objects</returns>
        public static Polynomial operator *(Polynomial p, double c)
        {
            Polynomial temp = new Polynomial();
            foreach (MultiVarTerm m1 in p._terms)
                temp.Add(m1 * c);
            return temp;
        }

        /// <summary>
        /// Multiplies a Polynomial with a double number
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The product of the two objects</returns>
        public static Polynomial operator *(double c, Polynomial p)
        {
            return p * c;
        }

        /// <summary>
        /// Adds a Polynomial with a Polynomial
        /// </summary>
        /// <param name="p1">The first Polynomial</param>
        /// <param name="p2">The second Polynomial</param>
        /// <returns>The sum of the two objects</returns>
        public static Polynomial operator +(Polynomial p1, Polynomial p2)
        {
            Polynomial temp = new Polynomial();
            foreach (MultiVarTerm m in p1._terms) //Adds the terms of p1
                temp.Add(m.Copy());
            foreach (MultiVarTerm m in p2._terms) //Adds the terms of p2
                temp.Add(m.Copy());
            return temp;
        }

        /// <summary>
        /// Adds a Polynomial with a MultiVarTerm
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="t">The MultiVarTerm</param>
        /// <returns>The sum of the two objects</returns>
        public static Polynomial operator +(Polynomial p, MultiVarTerm t)
        {
            Polynomial temp = new Polynomial();
            foreach (MultiVarTerm m in p._terms)
                temp.Add(m.Copy());
            temp.Add(t.Copy());
            return temp;
        }

        /// <summary>
        /// Adds a Polynomial with a MultiVarTerm
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="t">The MultiVarTerm</param>
        /// <returns>The sum of the two objects</returns>
        public static Polynomial operator +(MultiVarTerm t, Polynomial p)
        {
            return p + t;
        }

        /// <summary>
        /// Adds a Polynomial with a double number
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The sum of the two objects</returns>
        public static Polynomial operator +(Polynomial p, double c)
        {
            Polynomial temp = new Polynomial();
            foreach (MultiVarTerm m in p._terms)
                temp.Add(m.Copy());
            temp.Add(new MultiVarTerm(c));
            return temp;
        }

        /// <summary>
        /// Adds a Polynomial with a double number
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The sum of the two objects</returns>
        public static Polynomial operator +(double c, Polynomial p)
        {
            return p + c;
        }

        /// <summary>
        /// Calculates the opposite Polynomial
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <returns>The opposite Polynomial</returns>
        public static Polynomial operator -(Polynomial p)
        {
            Polynomial temp = new Polynomial();
            foreach (MultiVarTerm t in p._terms) //Calculates the opposite terms
                temp.Add(-t.Copy());
            return temp;
        }

        /// <summary>
        /// Substracts a Polynomial with a Polynomial
        /// </summary>
        /// <param name="p1">The first Polynomial</param>
        /// <param name="p2">The second Polynomial</param>
        /// <returns>The difference of the two objects</returns>
        public static Polynomial operator -(Polynomial p1, Polynomial p2)
        {
            return p1 + (-p2);
        }

        /// <summary>
        /// Substracts a Polynomial with a MultiVarTerm
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="t">The MultiVarTerm</param>
        /// <returns>The difference of the two objects</returns>
        public static Polynomial operator -(Polynomial p, MultiVarTerm t)
        {
            return p + (-t);
        }

        /// <summary>
        /// Substracts a Polynomial with a MultiVarTerm
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="t">The MultiVarTerm</param>
        /// <returns>The difference of the two objects</returns>
        public static Polynomial operator -(MultiVarTerm t, Polynomial p)
        {
            return - p + t;
        }

        /// <summary>
        /// Substracts a Polynomial with a double number
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The difference of the two objects</returns>
        public static Polynomial operator -(Polynomial p1, double c)
        {
            return p1 + (-c);
        }

        /// <summary>
        /// Substracts a Polynomial with a double number
        /// </summary>
        /// <param name="p">The Polynomial</param>
        /// <param name="c">The double number</param>
        /// <returns>The difference of the two objects</returns>
        public static Polynomial operator -(double c, Polynomial p)
        {
            return p - c;
        }
        #endregion

        #region Other Methods
        /// <summary>
        /// Returns the Polynomial in the form of a string
        /// </summary>
        /// <returns>A string</returns>
        public override string ToString()
        {
            Sort();
            StringBuilder str = new StringBuilder();

            bool first = true;
            for (int i = 0; i < _terms.Count; i++)
            {
                if (Math.Sign(_terms[i].Coefficient) == 1 && !first) { str.Append(" + "); }
                first = false;
                str.Append(_terms[i].ToString());
            }

            return str.ToString();
        }

        /// <summary>
        /// Returns the Polynomial in the form of a string
        /// </summary>
        /// <returns>A string</returns>
        public string ToStringMaple()
        {
            Sort();
            StringBuilder str = new StringBuilder();

            bool first = true;
            for (int i = 0; i < _terms.Count; i++)
            {
                if (Math.Sign(_terms[i].Coefficient) == 1 && !first) { str.Append(" + "); }
                first = false;
                str.Append(_terms[i].ToStringMaple());
            }
            str.Replace(',', '.');
            return str.ToString();
        }

        /// <summary>
        /// Creates a deep copy of the Polynomial
        /// </summary>
        /// <returns>A copy of the Polynomial</returns>
        public Polynomial Copy()
        {
            Polynomial copy = new Polynomial();

            foreach (MultiVarTerm term in this._terms)
                copy._terms.Add(term.Copy());
            //copy._terms.Sort();
            return copy;
        }

        /// <summary>
        /// Sorts the terms of the current Polynomial
        /// </summary>
        public void Sort()
        {
            _terms.Sort();
        }

        /// <summary>
        /// Sets the read-only property of the Polynomial
        /// </summary>
        /// <param name="readOnly">A boolean value determining whether the Polynomial will be read-only</param>
        public void SetReadOnly(bool readOnly)
        {
            _isReadOnly = readOnly;
        }
        #endregion
    }
}
