﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using PolyLib;

namespace Braids
{
    /// <summary>
    /// A class modeling a Word for Hecke and Yokonuma-Hecke algebras.
    /// </summary>
    public class Word
    {
        #region Properties
        /// <summary>
        /// The framing part of the word (if available)
        /// </summary>
        public List<int[]> Framing { get; set; }

        /// <summary>
        /// The braiding part of the word
        /// </summary>
        public List<int[]> Braiding { get; set; }

        /// <summary>
        /// The polynomial coefficient of the word
        /// </summary>
        public Polynomial Coeff { get; set; }
        #endregion

        #region Constructors
        /// <summary>
        /// Creates a unitary Word (1)
        /// </summary>
        public Word()
        {
            Framing = new List<int[]>();
            Braiding = new List<int[]>();
            Coeff = new Polynomial(1);
        }

        /// <summary>
        /// Creates a new Word
        /// </summary>
        /// <param name="elements">A list of integer arrays containing the elements of the word</param>
        public Word(List<int[]> elements)
        {
            Tuple<List<int[]>, List<int[]>> temp = MoveFramingPart(elements); //Separates the framing and braiding parts
            Framing = temp.Item1; //Sets the framing part
            Braiding = temp.Item2; //Sets the braiding part
            Coeff = new Polynomial(1); //Sets the default coefficient (1)
            RemoveDuplicates(); //Removes any duplicate elements
        }

        /// <summary>
        /// Creates a new Word
        /// </summary>
        /// <param name="elements">A list of integer arrays containing the elements of the word</param>
        /// <param name="coefficient">A polynomial coefficient</param>
        public Word(List<int[]> elements, Polynomial coefficient)
        {
            Tuple<List<int[]>, List<int[]>> temp = MoveFramingPart(elements); //Separates the framing and braiding parts
            Framing = temp.Item1; //Sets the framing part
            Braiding = temp.Item2; //Sets the braiding part
            Coeff = coefficient.Copy(); //Sets the coefficient
            RemoveDuplicates(); //Removes any duplicate elements
        }

        /// <summary>
        /// Creates a new Word
        /// </summary>
        /// <param name="framingPart">A list of integer arrays containing the framing part</param>
        /// <param name="braidingPart">A list of integer arrays containing the framing part</param>
        /// <param name="coefficient">A polynomial coefficient</param>
        public Word(List<int[]> framingPart, List<int[]> braidingPart, Polynomial coefficient)
        {
            //Checks if the integer arrays are valid
            foreach (int[] arr in framingPart)
                if (arr.Length < 2) { throw new Exception("Data are missing from an framing element"); }
            foreach (int[] arr in braidingPart)
                if (arr.Length < 2) { throw new Exception("Data are missing from an braiding element"); }

            //Sets the framing part
            Framing = new List<int[]>();
            for (int i = 0; i < framingPart.Count; i++)
                Framing.Add(new int[] {framingPart[i][0], framingPart[i][1]});

            //Sets the braiding part
            Braiding = new List<int[]>();
            for (int i = 0; i < braidingPart.Count; i++)
                Braiding.Add(new int[] {braidingPart[i][0], braidingPart[i][1]});

            Coeff = coefficient.Copy(); //Sets the coefficient
            RemoveDuplicates(); //Removes any duplicate elements
        }
        #endregion

        #region Operators
        /// <summary>
        /// Multiplies two Words by multiplying the coefficients and merging the elements
        /// </summary>
        /// <param name="w1">The first Word</param>
        /// <param name="w2">The second Word</param>
        /// <returns>A Word object (the product of the two words)</returns>
        public static Word operator *(Word w1, Word w2)
        {
            if (w2.Framing.Count == 0 && w2.Braiding.Count == 0) //w2 is an empty word
                return new Word(w1.Framing, w1.Braiding, w1.Coeff * w2.Coeff);
            else if (w1.Framing.Count == 0 && w1.Braiding.Count == 0) //w1 is an empty word
                return new Word(w2.Framing, w2.Braiding, w1.Coeff * w2.Coeff);
            else if (w2.Framing.Count == 0) //w2 has no framing
            {
                //Copies the braiding parts of both words to a new list
                List<int[]> braid = new List<int[]>();
                for (int i = 0; i < w1.Braiding.Count; i++)
                    braid.Add(new int[] { w1.Braiding[i][0], w1.Braiding[i][1] });
                for (int i = 0; i < w2.Braiding.Count; i++)
                    braid.Add(new int[] { w2.Braiding[i][0], w2.Braiding[i][1] });
                //Creates the new word
                return new Word(w1.Framing, braid, w1.Coeff * w2.Coeff);
            }
            else //w2 has framing and braiding part
            {
                //Copies the elements of the first word to the new list
                List<int[]> temp = new List<int[]>();
                for (int i = 0; i < w1.Framing.Count; i++)
                    temp.Add(new int[] { 't', w1.Framing[i][0], w1.Framing[i][1] });
                for (int i = 0; i < w1.Braiding.Count; i++)
                    temp.Add(new int[] { 'g', w1.Braiding[i][0], w1.Braiding[i][1] });
                //Copies the elements of the first word to the new list
                for (int i = 0; i < w2.Framing.Count; i++)
                    temp.Add(new int[] { 't', w2.Framing[i][0], w2.Framing[i][1] });
                for (int i = 0; i < w2.Braiding.Count; i++)
                    temp.Add(new int[] { 'g', w2.Braiding[i][0], w2.Braiding[i][1] });
                return new Word(temp, w1.Coeff * w2.Coeff);
            }
        }

        /// <summary>
        /// Multiplies a word with a polynomial
        /// </summary>
        /// <param name="w">The word</param>
        /// <param name="p">The polynomial</param>
        /// <returns>A Word object (the product)</returns>
        public static Word operator *(Word w, Polynomial p)
        {
            return new Word(w.Framing, w.Braiding, w.Coeff * p);
        }

        /// <summary>
        /// Multiplies a word with a polynomial
        /// </summary>
        /// <param name="p">The polynomial</param>
        /// <param name="w">The word</param>
        /// <returns>A Word object (the product)</returns>
        public static Word operator *(Polynomial p, Word w)
        {
            return w * p;
        }

        /// <summary>
        /// Multiplies a word with an element
        /// </summary>
        /// <param name="w1">The word</param>
        /// <param name="w2">The element</param>
        /// <returns>A Word object (the product)</returns>
        public static Word operator *(Word w, int[] elem)
        {
            if (elem.Length < 3) { throw new Exception("Data are missing from an element"); }

            if (elem[0] == 'g') //elem belongs to the braiding part
            {
                //Creates a list that contains the braiding part plus elem
                List<int[]> braiding = new List<int[]>();
                for (int i = 0; i < w.Braiding.Count; i++)
                    braiding.Add(w.Braiding[i]);
                braiding.Add(new int[] { elem[1], elem[2] });
                return new Word(w.Framing, braiding, w.Coeff);
            }
            else
            {
                //Creates a list that contains both braiding and framing pars
                List<int[]> temp = new List<int[]>();
                for (int i = 0; i < w.Framing.Count; i++)
                    temp.Add(new int[] { 't', w.Framing[i][0], w.Framing[i][1] });
                for (int i = 0; i < w.Braiding.Count; i++)
                    temp.Add(new int[] { 'g', w.Braiding[i][0], w.Braiding[i][1] });
                temp.Add(elem);
                return new Word(temp, w.Coeff);
            }
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Searches and applies all possible commutes
        /// </summary>
        /// <param name="maxIndexes">An array containing the indexes of the symbols with maximum index</param>
        private void SearchCommute(int[] maxIndexes, bool yokonuma)
        {
            DateTime dt1 = DateTime.Now;
            List<int[]> dist = new List<int[]>();

            for (int i = 0; i < maxIndexes.Length; i++)
            {
                for (int j = i; j < maxIndexes.Length; j++)
                    if (i != j)
                    {
                        if (maxIndexes[i] < maxIndexes[j]) { dist.Add(new int[] { maxIndexes[i], maxIndexes[j], maxIndexes[j] - maxIndexes[i] - 1 }); }
                        else { dist.Add(new int[] { maxIndexes[i], maxIndexes[j], Braiding.Count - (maxIndexes[i] - maxIndexes[j] - 1) - 2 }); }
                    }
            }
            dist.Sort(new DistComparer());
            dist.Reverse();
            if (dist.Count > 0) { Commute(dist[0][0], dist[0][1], false, 0, yokonuma); }
            RemoveDuplicates();
            DateTime dt2 = DateTime.Now;
            Time.Milliseconds += (dt2 - dt1).TotalMilliseconds;
        }

        /// <summary>
        /// Applies all possible commutative relations for the item in <paramref name="startIndex"/>
        /// </summary>
        /// <param name="Elements">A Elements with no powers of gi from <paramref name="startIndex"/> to <paramref name="endIndex"/></param>
        /// <param name="startIndex">The start</param>
        /// <param name="endIndex">The end</param>
        /// <param name="ignoreExp">If true, it ignores any exponent -> returns always false</param>
        /// <param name="offset">An offset (for future use)</param>
        /// <returns>A boolean value describing whether the Elements contains a power</returns>
        private bool Commute(int startIndex, int endIndex, bool ignoreExp, int offset, bool yokonuma)
        {
            int newPos = -1; //The new position of the item
            //TODO: int max = 

            if (startIndex < offset)
            {
                throw new Exception();
            }
            if (Braiding.Count - endIndex - 1 < offset)
            {
                throw new Exception();
            }

            //Moves start item from left to right
            int oldCount = Braiding.Count; //Word length before any move
            int posL = BubbleHecke(startIndex, endIndex, offset, yokonuma); //The new startIndex
            int newCount = Braiding.Count; //Word length after

            if (ignoreExp)
            {
                if (oldCount - newCount > 0)
                    endIndex -= oldCount - newCount; //endIndex has been moved
            }
            else
            {
                if (Braiding[posL][1] > 1)
                    return true; //The item has an exponent larger than 1
            }

            if (Braiding[posL][1] > 1 && !ignoreExp)
                return true; //The item has an exponent larger than 1 ---> replace!

            //Moves end item from right to left
            int posR = BubbleHecke(endIndex, startIndex, offset, yokonuma); //The new endIndex
            if (Braiding[posR][1] > 1 && !ignoreExp)
                return true; //The item has an exponent larger than 1 ---> replace!

            int middle = posL + (posR - posL) / 2; //The middle index

            //Commute from middle to left
            for (int i = posL + 1; i <= posR; i++) //Middle will be moved too
            {
                newPos = BubbleHecke(i, posL, offset, yokonuma);
                if (i == posR && newPos < posR) { continue; }
                bool beforeWord = newPos == posL; //The item is moved before the start item
                bool exp = false;
                if (newPos > -1) { exp = Braiding[newPos][1] > 1; } //Checks if an exponent greater than 1 is created (BubbleHecke automatically finds squares)

                if (beforeWord)
                {
                    if (exp && !ignoreExp)
                        return true; //(iii) Square found by BubbleHecke
                    else
                    {
                        posL++; //(i) The index of the left item is moved
                        middle = posL + (posR - posL) / 2; //Middle must be recalculated
                    }
                }
                else
                    if (exp && !ignoreExp)
                        return true; //(ii) Square found by BubbleHecke
            }

            //Commute from middle to right
            //for (int i = posL + 1; i < posR; i++)
            //{
            //    newPos = BubbleHecke(i, posR, offset, yokonuma);
            //    bool afterWord = newPos == posR - 1; //The item is moved before the start item
            //    bool exp = Braiding[newPos][1] > 1; //Checks if an exponent greater than 1 is created (BubbleHecke automatically finds squares)

            //    if (afterWord)
            //    {
            //        if (exp && !ignoreExp)
            //            return true; //(iii) Square found by BubbleHecke
            //        else
            //        {
            //            posR--; //(i) The index of the right item is moved
            //            middle = posL + (posR - posL) / 2; //Middle must be recalculated
            //        }
            //    }
            //    else
            //        if (exp && !ignoreExp)
            //            return true; //(ii) Square found by BubbleHecke
            //}

            return false; //No powers, all are OK
        }

        /// <summary>
        /// Makes all possible commutes in braiding part for the item located at <paramref name="start"/>
        /// </summary>
        /// <param name="start">The start index</param>
        /// <param name="end">The end index</param>
        /// <returns>The new position of the item</returns>
        private int BubbleHecke(int start, int end, int offset, bool yokonuma)
        {
            //Console.WriteLine(ToString());
            if (start < end)
            {
                int pos = start; //The new position of the item
                do
                    pos++;
                while (pos <= end && Math.Abs(Braiding[pos][0] - Braiding[start][0]) > 1); //Checks for all possible commutes (far commutativity)
                pos--; //because of the while loop

                for (int i = 0; i <= offset; i++)
                    Swap(Braiding, start - i, pos - i); //Moves the item

                //Checks for duplicates
                //if ((pos < Braiding.Count - 1 || !yokonuma) && Braiding[pos][0] == Braiding[Compute.Mod((pos + 1), Braiding.Count)][0])
                //{
                //    Braiding[pos][1] += Braiding[Compute.Mod((pos + 1), Braiding.Count)][1];
                //    Braiding.RemoveAt((pos + 1) % Braiding.Count);
                //    if (pos == end) //end does not exist any more (duplicate was removed)
                //        return pos - 1;
                //}
                //if (pos < Braiding.Count && (!yokonuma || (pos > 0 && yokonuma)) && Braiding[pos][0] == Braiding[Compute.Mod((pos - 1), Braiding.Count)][0])
                //{
                //    Braiding[Compute.Mod((pos - 1), Braiding.Count)][1] += Braiding[pos][1];
                //    Braiding.RemoveAt(pos);
                //    return pos - 1; //Returns the new position of the item
                //}
                if (Braiding[pos][0] == Braiding[Compute.Mod((pos + 1), Braiding.Count)][0])
                {
                    Braiding[pos][1] += Braiding[Compute.Mod((pos + 1), Braiding.Count)][1];
                    Braiding.RemoveAt((pos + 1) % Braiding.Count);
                    if (pos == end) //end does not exist any more (duplicate was removed)
                        return pos - 1;
                }
                if (!yokonuma && Braiding[pos][0] == Braiding[Compute.Mod((pos - 1), Braiding.Count)][0])
                {
                    Braiding[Compute.Mod((pos - 1), Braiding.Count)][1] += Braiding[pos][1];
                    Braiding.RemoveAt(pos);
                    return pos - 1; //Returns the new position of the item
                }
                return pos; //Returns the new position of the item
            }
            else if (start == end) //No commute needed
                return start; //Returns the start index
            else
            {
                int pos = start; //The new position of the item
                do
                    pos--;
                while (pos >= end && Math.Abs(Braiding[pos][0] - Braiding[start][0]) > 1); //Checks for all possible commutes (far commutativity)
                pos++; //because of the while loop

                for (int i = 0; i <= offset; i++)
                    Swap(Braiding, start + i, pos + i); //Moves the item

                //Checks for duplicates
                if (Braiding[Compute.Mod((pos - 1), Braiding.Count)][0] == Braiding[pos][0])
                {
                    Braiding[Compute.Mod((pos - 1), Braiding.Count)][1] += Braiding[pos][1];
                    Braiding.RemoveAt(pos);
                    return pos - 1;
                }
                if (!yokonuma && Braiding[pos][0] == Braiding[Compute.Mod((pos + 1), Braiding.Count)][0])
                {
                    Braiding[pos][1] += Braiding[Compute.Mod((pos + 1), Braiding.Count)][1];
                    Braiding.RemoveAt(Compute.Mod((pos + 1), Braiding.Count));
                }
                return pos; //Returns the new position of the item
            }
        }

        /// <summary>
        /// Swaps <paramref name="data1"/> and <paramref name="data2"/> (and the items between) in the <paramref name="part"/>
        /// </summary>
        /// <param name="part">The framing or the braiding part</param>
        /// <param name="data1">The first item</param>
        /// <param name="data2">The second item</param>
        private void Swap(List<int[]> part, int data1, int data2)
        {
            if (data1 < data2) //Swap to left ( i <- i + 1)
            {
                int[] temp = part[data1];
                //Moves all data to the left
                for (int i = data1; i < data2; i++)
                    part[i] = part[i + 1];
                //Moves data1 to data2
                part[data2] = temp;
            }
            else if (data1 > data2) //Swap to right ( i - 1 -> i)
            {
                int[] temp = part[data1];
                //Moves all data to the right
                for (int i = data1; i > data2; i--)
                    part[i] = part[i - 1];
                //Moves data1 to data2
                part[data2] = temp;
            }
            else //The two items are the same -> no swap needed
                return;
        }

        /// <summary>
        /// Checks if the <paramref name="part"/> contains any power
        /// </summary>
        /// <param name="part">The framing or the braiding part</param>
        /// <returns>A boolean value determining whether the <paramref name="part"/> contains any exponent != 1</returns>
        private bool HasPowers(List<int[]> part)
        {
            return ((from w in part where w[1] != 1 select w).Count() > 0);
        }

        /// <summary>
        /// Returns the indexes of the <paramref name="part"/> are raised to a power
        /// </summary>
        /// <param name="part">The framing or the braiding part</param>
        /// <returns>An array of integer containing the indexes</returns>
        private int[] FindPowers(List<int[]> part)
        
        {
            List<int> _indexes = new List<int>();
            List<int> _powers = new List<int>();
            RemoveDuplicates(); //Removes possible zero exponents
            for (int i = 0; i < part.Count; i++)
            {
                if (part[i][1] != 1)
                {
                    _indexes.Add(i);
                    _powers.Add(part[i][1]);
                }
            }
            int[] indexes = _indexes.ToArray<int>();
            int[] powers = _powers.ToArray<int>();
            //int[] temp = (from e in part where e[1] != 1 select part.IndexOf(e)).ToArray<int>();
            Array.Sort(powers, indexes, new AbsComparer());
            return indexes;
        }

        /// <summary>
        /// Finds the indexes with the maximum subscript in <paramref name="part"/>
        /// </summary>
        /// <param name="part">The framing or the braiding part</param>
        /// <returns>An array of integers containing the indexes</returns>
        public int[] FindMaxIndexes(List<int[]> part)
        {
            List<int> index = new List<int>();
            List<int> subscr = (from w in part select w[0]).ToList<int>();
            for (int i = 0; i < part.Count; i++)
                if (subscr[i] == subscr.Max())
                    index.Add(i);
            return index.ToArray<int>();
        }

        public int[] FindIndexes(List<int[]> part, int ind)
        {
            List<int> index = new List<int>();
            List<int> subscr = (from w in part select w[0]).ToList<int>();
            for (int i = 0; i < part.Count; i++)
                if (subscr[i] == ind)
                    index.Add(i);
            return index.ToArray<int>();
        }

        public int ExpSum()
        {
            int sum =0;
            for (int i = 0; i < Braiding.Count; i++)
                sum += Braiding[i][1];
            return sum;
        }

        /// <summary>
        /// Checks if the braiding part contains a brigde
        /// </summary>
        /// <returns>An integer tuple with 2 items, containing the start and the end of the bridge</returns>
        private Tuple<int, int> FindBridge()
        {
            List<int> localMinimaList = new List<int>();

            for (int i = 1; i < Braiding.Count - 1; i++)
                if (Braiding[i][0] == Braiding[i + 1][0] - 1 && Braiding[i][0] == Braiding[i - 1][0] - 1)
                    localMinimaList.Add(i);

            int[] localMinima = localMinimaList.ToArray<int>();
            int count = localMinima.Length; //Count of local minima

            //Check if local minima exist
            //If there isn't any local minimum, the Elements must contain a power, so we shouldn't have called this method
            //But I check for the existence only for consistency
            if (count > 0)
            {
                int[] bridgeSize = new int[count]; //Contains the size of every bridge found for each local minimum

                //Fills the list with zeros
                for (int i = 0; i < count; i++)
                    bridgeSize[i] = 0;

                int dist; //Dump variable ---> helps in checking the
                for (int i = 0; i < count; i++)
                {
                    dist = 1;
                    while (Braiding[Compute.Mod(localMinima[i] - dist, Braiding.Count)][0] == Braiding[Compute.Mod(localMinima[i] + dist, Braiding.Count)][0] //The left and the right elements are the same
                        && Braiding[Compute.Mod(localMinima[i] - dist, Braiding.Count)][0] == Braiding[Compute.Mod(localMinima[i] - dist + 1, Braiding.Count)][0] + 1)//+1 from the previous elements
                    {
                        dist++;
                        bridgeSize[i]++;
                    }
                }

                Array.Sort(bridgeSize, localMinima); //Sorts the arrays with descending sorting based on brigdeSize array

                return new Tuple<int, int>(Compute.Mod(localMinima[count - 1] - bridgeSize[count - 1], Braiding.Count), Compute.Mod(localMinima[count - 1] + bridgeSize[count - 1], Braiding.Count));
            }


            return new Tuple<int, int>(-1, -1);
        }

        /// <summary>
        /// Replaces a bridge with a mountain
        /// </summary>
        /// <param name="start">The start of the bridge</param>
        /// <param name="end">The end of the bridge</param>
        private void ReplaceBridge(int start, int end)
        {
            if (start < end) //Internal bridge
            {
                int half = (end - start) / 2; //The middle index
                int max = Braiding[start][0]; //The maximum item
                int min = Braiding[start + half][0]; //The minimum item

                for (int i = 0; i < half; i++) //Replaces the items
                {
                    //int[] temp = new int[] { min + i, 1 }; //The new item
                    Braiding[start + i] = new int[] { min + i, 1 }; //Replace of the left item
                    Braiding[end - i] = new int[] { min + i, 1 }; //Replace of the right item
                }
                Braiding[start + half] = new int[] { max, 1 }; //Replace of the middle item
            }
            else if (end < start) //External bridge
            {
                int[] temp;
                //Swap tr(ab) = tr(ba)

                if (end <= Braiding.Count / 2) //Swaps from the end to the beginning of the Elements
                {
                    for (int i = 0; i < Braiding.Count - start; i++)
                    {
                        temp = new int[] { Braiding[Braiding.Count - 1][0], Braiding[Braiding.Count - 1][1] };
                        Braiding.RemoveAt(Braiding.Count - 1);
                        Braiding.Insert(0, temp);
                    }
                    ReplaceBridge(0, end + Braiding.Count - start); //Replaces the new internal bridge
                }
                else //Swaps from the beginning to the end of the Elements
                {
                    for (int i = 0; i <= end; i++)
                    {
                        temp = new int[] {Braiding[0][0], Braiding[0][1]};
                        Braiding.RemoveAt(0);
                        Braiding.Add(temp);
                    }
                    ReplaceBridge(start - end - 1, Braiding.Count - 1); //Replaces the new internal bridge
                }

            }
            RemoveDuplicates();
        }

        /// <summary>
        /// Removes duplicates from both framing and braiding parts
        /// </summary>
        private void RemoveDuplicates()
        {
            Framing.Sort(new FramingComparer());
            Framing.Reverse();

            for (int i = 0; i < Framing.Count - 1; i++)
                if (Framing[i][0] == Framing[i + 1][0])
                {
                    Framing[i][1] += Framing[i + 1][1];
                    Framing.RemoveAt(i + 1);
                    i--;
                }
            if (Framing.Count > 1 && Framing[0][0] == Framing[Framing.Count - 1][0]) //Checks if the first and the last item have the same index (count = 1 <=> start = end)
            {
                Framing[0][1] += Framing[Framing.Count - 1][1];
                Framing.RemoveAt(Braiding.Count - 1);
            }

            //for (int i = 0; i < Braiding.Count - 1; i++)
            //{
            //    for (int j = i + 1; j < Braiding.Count; j++)
            //        BubbleHecke(i, j, 0, false);
            //}
            
            for (int i = 0; i < Braiding.Count - 1; i++)
                if (Braiding[i][0] == Braiding[i + 1][0])
                {
                    Braiding[i][1] += Braiding[i + 1][1];
                    Braiding.RemoveAt(i + 1);
                    i--;
                }

            if (Framing.Count == 0 && Braiding.Count > 1 && Braiding[0][0] == Braiding[Braiding.Count - 1][0]) //Checks if the first and the last item have the same index (count = 1 <=> start = end)
            {
                Braiding[0][1] += Braiding[Braiding.Count - 1][1];
                Braiding.RemoveAt(Braiding.Count - 1);
            }

            DeleteZeroExp();
        }

        /// <summary>
        /// Makes all apporopriate calculations (mod d) at the exponents of the framing part
        /// </summary>
        /// <param name="d">The integer d of Y_d,n(u)</param>
        private void RemoveModDuplicates(int d)
        {
            for (int i = 0; i < Framing.Count; i++)
                Framing[i][1] = Compute.Mod(Framing[i][1], d);
        }

        /// <summary>
        /// Deletes all the elements with zero exponent in the framing part
        /// </summary>
        private void DeleteZeroExp()
        {
            for (int i = 0; i < Framing.Count; i++)
                if (Framing[i][1] == 0)
                {
                    Framing.RemoveAt(i);
                    i--;
                }
            for (int i = 0; i < Braiding.Count; i++)
                if (Braiding[i][1] == 0)
                {
                    Braiding.RemoveAt(i);
                    i--;
                }
        }

        //TODO: Update comments
        private Tuple<List<int[]>, List<int[]>> Split(int pos)
        {            
            List<int[]> leftPart = new List<int[]>();
            List<int[]> rightPart = new List<int[]>();
            
            for (int i = 0; i < pos; i++)
                leftPart.Add(new int[] {Braiding[i][0], Braiding[i][1] });
            for (int i=pos+1; i < Braiding.Count; i++)
                rightPart.Add(new int[] {Braiding[i][0], Braiding[i][1] });

            return new Tuple<List<int[]>, List<int[]>>(leftPart, rightPart) ;
        }

        private Tuple<List<int[]>, List<int[]>> SplitYokonuma(int pos)
        {
            List<int[]> leftPart = new List<int[]>();
            List<int[]> rightPart = new List<int[]>();

            for (int i = 0; i < Framing.Count; i++)
                leftPart.Add(new int[] { 't', Framing[i][0], Framing[i][1] });
            for (int i = 0; i < pos; i++)
                leftPart.Add(new int[] { 'g', Braiding[i][0], Braiding[i][1] });
            for (int i = pos + 1; i < Braiding.Count; i++)
                rightPart.Add(new int[] { 'g', Braiding[i][0], Braiding[i][1] });

            return new Tuple<List<int[]>, List<int[]>>(leftPart, rightPart);
        }


        /// <summary>
        /// Splits the word to two different parts (the coefficient is copied only to the first word)
        /// </summary>
        /// <param name="frame">True if <paramref name="pos"/> is in the framing part, false for the braiding part</param>
        /// <param name="pos">The position of the split in <paramref name="part"/></param>
        /// <returns>An array of Words containing the left and the right parts</returns>
        [Obsolete]
        private Word[] Split_old(bool frame, int pos)
        {
            Word[] temp = new Word[2];
            temp[0] = new Word();
            //temp[0].Coeff = Coeff;
            temp[1] = new Word();

            if (frame)
            {
                //Left part
                for (int i = 0; i < pos; i++)
                    temp[0].Framing.Add(new int[] { Framing[i][0], Framing[i][1] });
                //Right part
                for (int i = pos + 1; i < Framing.Count; i++)
                    temp[1].Framing.Add(new int[] { Framing[i][0], Framing[i][1] });
                for (int i = 0; i < Braiding.Count; i++)
                    temp[1].Braiding.Add(new int[] { Braiding[i][0], Braiding[i][1] });
            }
            else
            {
                //Left part
                for (int i = 0; i < Framing.Count; i++)
                    temp[0].Framing.Add(new int[] { Framing[i][0], Framing[i][1] });
                for (int i = 0; i < pos; i++)
                    temp[0].Braiding.Add(new int[] { Braiding[i][0], Braiding[i][1] });
                //Right part
                for (int i = pos + 1; i < Braiding.Count; i++)
                    temp[1].Braiding.Add(new int[] { Braiding[i][0], Braiding[i][1] });
            }
            temp[0].RemoveDuplicates();
            //temp[1].RemoveDuplicates();
            return temp;
        }

        /// <summary>
        /// Moves the framing part to the start of the word
        /// </summary>
        /// <param name="elements">A list containing all the elements of the word</param>
        /// <returns>A Tuple containing the framing and the braiding parts separately</returns>
        private static Tuple<List<int[]>, List<int[]>> MoveFramingPart(List<int[]> elements)
        {
            foreach (int[] arr in elements)
                if (arr.Length < 3) { throw new Exception("Data are missing from an element"); }

            List<int[]> framing = new List<int[]>();
            List<int[]> braiding = new List<int[]>();
            for (int i = 0; i < elements.Count; i++)
            {
                if (elements[i][0] == 't')
                {
                    int result = elements[i][1];
                    //σ_i(j)
                    for (int j = i - 1; j > -1; j--)
                    {
                        if (elements[j][1] == result && elements[j][0] == 'g' && elements[j][2] % 2 != 0)
                            result++;
                        else if (elements[j][1] + 1 == result && elements[j][0] == 'g' && elements[j][2] % 2 != 0)
                            result--;
                    }
                    elements[i][1] = result;
                    framing.Add(new int[] { elements[i][1], elements[i][2] });
                    elements.RemoveAt(i);
                    i--;
                }
            }

            for (int i = 0; i < elements.Count; i++)
                braiding.Add(new int[] { elements[i][1], elements[i][2] });

            return new Tuple<List<int[]>, List<int[]>>(framing, braiding);
        }

        /// <summary>
        /// Multiplies two Words by multiplying the coefficients and merging the elements
        /// </summary>
        /// <param name="w1">The first Word</param>
        /// <param name="w2">The second Word</param>
        /// <returns>A Word object (the product of the two words)</returns>
        private static List<int[]> MergeHecke(Word w1, int[] elem)
        {
            //Copies the braiding parts of both words to a new list
            List<int[]> braid = new List<int[]>();
            for (int i = 0; i < w1.Braiding.Count; i++)
                braid.Add(new int[] { w1.Braiding[i][0], w1.Braiding[i][1] });
            braid.Add(elem);
            return braid;
        }

        /// <summary>
        /// Multiplies two Words by multiplying the coefficients and merging the elements
        /// </summary>
        /// <param name="w1">The first Word</param>
        /// <param name="w2">The second Word</param>
        /// <returns>A Word object (the product of the two words)</returns>
        private static List<int[]> MergeHeckeWords(Word w1, Word w2)
        {
            //Copies the braiding parts of both words to a new list
            List<int[]> braid = new List<int[]>();
            for (int i = 0; i < w1.Braiding.Count; i++)
                braid.Add(new int[] { w1.Braiding[i][0], w1.Braiding[i][1] });
            for (int i = 0; i < w2.Braiding.Count; i++)
                braid.Add(new int[] { w2.Braiding[i][0], w2.Braiding[i][1] });
            return braid;
        }
        #endregion

        #region Trace Calculation
        /// <summary>
        /// Calculates the Ocneanu trace for the braiding part.
        /// </summary>
        /// <returns>A complex polynomial.</returns>
        public Polynomial TraceHecke()
        {
            Framing.Clear();
            RemoveDuplicates();

            if (Braiding.Count == 0)
                return Coeff;
            if (Braiding.Count == 1)
                return Coeff * TraceHeckeUnit(Braiding[0]);
            else
            {
                int[] maxExp = FindPowers(Braiding);
                if (maxExp.Length > 0) //The Elements contains powers
                {
                    //for (int i = 0; i < maxExp.Length; i++)
                    //{
                        List<Word> words = QuadraticRelationHecke(maxExp[0]);
                        Polynomial poly = Coeff * (words[0].TraceHecke() + words[1].TraceHecke());
                        return poly;
                    //}
                }
                else //No powers in Elements
                {
                    //Console.WriteLine(ToString());
                    int[] maxInd = FindMaxIndexes(Braiding); //Finds the indexes of maximum subscripts
                    if (maxInd.Length == 1) //Only one maximum subscript
                    {
                        Braiding.RemoveAt(maxInd[0]); //Removes the letter containing the maximum subscript
                        Polynomial poly = TraceHecke();
                        return new Polynomial(new MultiVarTerm(1, 'z', 1)) * poly; //z * tr(remaining Elements)
                    }
                    else //Multiple maximum subscripts
                    {
                        SearchCommute(maxInd, false);
                        maxExp = FindPowers(Braiding); //Re-calculates for powers
                        if (maxExp.Length > 0) //The Elements contains powers
                        {
                            for (int i = 0; i < maxExp.Length; i++)
                            {
                                List<Word> words = QuadraticRelationHecke(maxExp[0]);
                                Polynomial poly = Coeff * (words[0].TraceHecke() + words[1].TraceHecke());
                                return poly;
                            }
                        }
                        else //No powers after commute
                        {
                            Tuple<int, int> index = FindBridge(); //Searches for a bridge

                            int j=1;
                            int max = Braiding[maxInd[0]][0];
                            while (index.Item1 == -1 && j != max - 2)
                            {
                                SearchCommute(FindIndexes(Braiding, max - j),false);
                                maxExp = FindPowers(Braiding); //Re-calculates for powers
                                if (maxExp.Length > 0) //The Elements contains powers
                                {
                                    for (int i = 0; i < maxExp.Length; i++)
                                    {
                                        List<Word> words = QuadraticRelationHecke(maxExp[0]);
                                        Polynomial poly = Coeff * (words[0].TraceHecke() + words[1].TraceHecke());
                                        return poly;
                                    }
                                }
                                index = FindBridge();
                                j++;
                            }

                            if (index.Item1 == -1) //MUST NOT BE REACHED: if reached, there is a set of braids whose trace can not be calculated
                                throw new Exception("The trace for this word cannot be calculated"); //Exception thrown
                            else //Bridge found
                            {
                                Word newWord = new Word(Framing, Braiding, Coeff);
                                newWord.ReplaceBridge(index.Item1, index.Item2); //Replaces the bridge with a mountain
                                //Polynomial poly = TraceHecke();
                                return newWord.TraceHecke(); //Calls recursively the trace function
                            }
                        }
                    }
                }
            }
            
            return null; //TODO: Delete
        }

        /// <summary>
        /// Calculates the Ocneanu trace for a single element
        /// </summary>
        /// <param name="symbol">The element</param>
        /// <returns>A complex polynomial</returns>
        private static Polynomial TraceHeckeUnit(int[] symbol)
        {
            if (symbol[1] == 0)
                return new Polynomial(new MultiVarTerm(1));
            else if (symbol[1] == 1)
                return new Polynomial(new MultiVarTerm(1, new Tuple<char, int>[] { new Tuple<char, int>('z', 1) }));
            else if (symbol[1] > 1)
            {
                int sign = 1;
                List<MultiVarTerm> temp = new List<MultiVarTerm>();
                for (int i = Math.Abs(symbol[1]) - 1; i > 0; i--)
                {
                    temp.Add(new MultiVarTerm(sign, new Tuple<char, int>('q', Math.Sign(symbol[1]) * i)));
                    sign *= -1;
                }
                Polynomial poly = new Polynomial(temp);
                MultiVarTerm z = new MultiVarTerm(1, new Tuple<char, int>[] { new Tuple<char, int>('z', 1) });
                return (poly - (int)Math.Pow(-1, symbol[1])) * z + poly;
            }
            else //<0
            {
                int sign = 1;
                List<MultiVarTerm> temp = new List<MultiVarTerm>();
                for (int i = Math.Abs(symbol[1]); i > 0; i--)
                {
                    temp.Add(new MultiVarTerm(sign, new Tuple<char, int>('q', Math.Sign(symbol[1]) * i)));
                    sign *= -1;
                }
                Polynomial poly = new Polynomial(temp);
                MultiVarTerm z = new MultiVarTerm(1, new Tuple<char, int>[] { new Tuple<char, int>('z', 1) });
                return (poly * z + poly + (int)Math.Pow(-1, symbol[1]));
            }
        }

        /// <summary>
        /// Applies the Hecke quadratic relation for a word
        /// </summary>
        /// <param name="pos">The position of the letter in the braiding part to be removed and substituted from the trace</param>
        /// <returns>A list of words whose sum is equal to the original word</returns>
        //public List<Word> QuadraticRelationHecke(int pos)
        //{
        //    int pow = Braiding[pos][1]; //The power to which the specified element is raised
        //    int sign = 1; //Dump variable (used to construct the polynomail-coefficient)

        //    Tuple<List<int[]>, List<int[]>> split = Split(pos); //Splits the word to two parts

        //    List<int[]> braid1 = new List<int[]>();
        //    List<int[]> braid2 = new List<int[]>();

        //    int count1 = -1;
        //    int count2 = -1;

        //    if (split.Item1.Count > 0) 
        //    { 
        //        braid1.Add(new int[] { split.Item1[0][0], split.Item1[0][1] });
        //        count1++;
        //    }
        //    if (split.Item2.Count > 0)
        //    {
        //        braid2.Add(new int[] { split.Item2[0][0], split.Item2[0][1] });
        //        count2++;
        //    }

        //    for (int i = 1; i < split.Item1.Count; i++)
        //    {
        //        if (count1 > -1  && braid1[count1][0] == split.Item1[i][0])
        //        {
        //            if (braid1[count1][1] == -split.Item1[i][1])
        //            {
        //                braid1.RemoveAt(count1);
        //                count1--;
        //            }
        //            else
        //                braid1[count1][1] += split.Item1[i][1];
        //        }
        //        else
        //        {
        //            braid1.Add(new int[] { split.Item1[i][0], split.Item1[i][1] });
        //            count1++;
        //        }

        //        if (count2 > -1 && braid2[count2][0] == split.Item1[i][0])
        //        {
        //            if (braid2[count2][1] == -split.Item1[i][1])
        //            {
        //                braid2.RemoveAt(count2);
        //                count2--;
        //            }
        //            else
        //                braid2[count2][1] += split.Item1[i][1];
        //        }
        //        else
        //        {
        //            braid2.Add(new int[] { split.Item1[i][0], split.Item1[i][1] });
        //            count2++;
        //        }
        //    }
            
        //    braid1.Add(new int[] { Braiding[pos][0], 1 });
            
        //    for (int i = 0; i < split.Item2.Count; i++)
        //    {
        //        if (count1 > -1 && braid1[count1][0] == split.Item2[i][0])
        //        {
        //            if (braid1[count1][1] == -split.Item2[i][1])
        //            {
        //                braid1.RemoveAt(count1);
        //                count1--;
        //            }
        //            else
        //                braid1[count1][1] += split.Item2[i][1];
        //        }
        //        else
        //        {
        //            braid1.Add(new int[] { split.Item2[i][0], split.Item2[i][1] });
        //            count1++;
        //        }

        //        if (count2 > -1 && braid2[count2][0] == split.Item2[i][0])
        //        {
        //            if (braid2[count2][1] == -split.Item2[i][1])
        //            {
        //                braid2.RemoveAt(count2);
        //                count2--;
        //            }
        //            else
        //                braid2[count2][1] += split.Item2[i][1];
        //        }
        //        else
        //        {
        //            braid2.Add(new int[] { split.Item2[i][0], split.Item2[i][1] });
        //            count2++;
        //        }
        //    }

        //    Polynomial poly = new Polynomial();

        //    Word w1 = new Word(Framing, braid1, new Polynomial(1));
        //    Word w2 = new Word(Framing, braid2, new Polynomial(1));

        //    if (pow > 0)
        //    {
        //        //Constructs q^(n-1) - q^(n-2) + ... + (-1)^n*q polynomial
                
        //        for (int i = Math.Abs(pow) - 1; i > 0; i--)
        //        {
        //            poly.Add(new MultiVarTerm(sign, new Tuple<char, int>('q', Math.Sign(pow) * i)));  //q^(n-1) + ...
        //            sign *= -1;
        //        }

        //        w1.Coeff *= poly - (int)Math.Pow(-1, pow); //w1 = (C - (-1)^n) A gi B
        //        w2.Coeff *= poly; //w2 = C A B
        //    }
        //    else // pow != 0 always
        //    {
        //        //Constructs q^(n-1) - q^(n-2) + ... + (-1)^n*q polynomial
        //        for (int i = Math.Abs(pow); i > 0; i--)
        //        {
        //            poly.Add(new MultiVarTerm(sign, new Tuple<char, int>('q', Math.Sign(pow) * i)));  //q^(n-1) + ...
        //            sign *= -1;
        //        }

        //        w1.Coeff *= poly; //w1 = (C - (-1)^n) A gi B
        //        w2.Coeff *= poly + (int)Math.Pow(-1, pow); //w2 = C A B
        //    }
        //    List<Word> output = new List<Word>(); //Creates the list of the new words
        //    output.Add(w1);
        //    output.Add(w2);
        //    return output;
        //}
        public List<Word> QuadraticRelationHecke(int pos)
        {
            int pow = Braiding[pos][1]; //The power to which the specified element is raised
            int sign = 1; //Dump variable (used to construct the polynomail-coefficient)

            Tuple<List<int[]>, List<int[]>> split = Split(pos); //Splits the word to two parts

            List<int[]> braid1 = new List<int[]>();
            List<int[]> braid2 = new List<int[]>();

            for (int i = 0; i < split.Item1.Count; i++)
            {
                braid1.Add(new int[] { split.Item1[i][0], split.Item1[i][1] });
                braid2.Add(new int[] { split.Item1[i][0], split.Item1[i][1] });
            }
            braid1.Add(new int[] { Braiding[pos][0], 1 });
            for (int i = 0; i < split.Item2.Count; i++)
            {
                braid1.Add(new int[] { split.Item2[i][0], split.Item2[i][1] });
                braid2.Add(new int[] { split.Item2[i][0], split.Item2[i][1] });
            }

            Polynomial poly = new Polynomial();

            Word w1 = new Word(Framing, braid1, new Polynomial(1));
            Word w2 = new Word(Framing, braid2, new Polynomial(1));

            if (pow > 0)
            {
                //Constructs q^(n-1) - q^(n-2) + ... + (-1)^n*q polynomial

                for (int i = Math.Abs(pow) - 1; i > 0; i--)
                {
                    poly.Add(new MultiVarTerm(sign, new Tuple<char, int>('q', Math.Sign(pow) * i)));  //q^(n-1) + ...
                    sign *= -1;
                }

                w1.Coeff *= poly - (int)Math.Pow(-1, pow); //w1 = (C - (-1)^n) A gi B
                w2.Coeff *= poly; //w2 = C A B
            }
            else // pow != 0 always
            {
                //Constructs q^(n-1) - q^(n-2) + ... + (-1)^n*q polynomial
                for (int i = Math.Abs(pow); i > 0; i--)
                {
                    poly.Add(new MultiVarTerm(sign, new Tuple<char, int>('q', Math.Sign(pow) * i)));  //q^(n-1) + ...
                    sign *= -1;
                }

                w1.Coeff *= poly; //w1 = (C - (-1)^n) A gi B
                w2.Coeff *= poly + (int)Math.Pow(-1, pow); //w2 = C A B
            }
            List<Word> output = new List<Word>(); //Creates the list of the new words
            output.Add(w1);
            output.Add(w2);
            return output;
        }

        /// <summary>
        /// Calculates the Yokonuma-Hecke trace for simple words (one framing element and/or one braiding element with exp=1)
        /// </summary>
        /// <param name="framingPower">The exponent of the framing element</param>
        /// <param name="hasBraiding">A boolean determining whether there exists a braiding element</param>
        /// <returns></returns>
        private static Polynomial TraceYokonumaHeckeUnit(int framingPower, bool hasBraiding)
        {
            if (framingPower == 0)
            {
                if (!hasBraiding)
                    return new Polynomial(1);
                else
                    return new Polynomial(new MultiVarTerm(1, 'z', 1));
            }
            else
            {
                if (!hasBraiding)
                    return new Polynomial(new MultiVarTerm(1, new Tuple<char, int>((char)(64 + framingPower), 1))); //ASCII: 65 = 'A'
                else
                    return new Polynomial(new MultiVarTerm(1, 'z', 1), new MultiVarTerm(1, new Tuple<char, int>((char)(64 + framingPower), 1))); //ASCII: 65 = 'A'
            }
        }

        /// <summary>
        /// Applies the Yokonuma-Hecke quadratic relation for a word
        /// </summary>
        /// <param name="d">The integer d of Y_d,n(u)</param>
        /// <param name="pos">The position of the letter in the braiding part to be removed and substituted from the trace</param>
        /// <returns>A list of words whose sum is equal to the original word</returns>
        public List<Word> QuadraticRelationYokonumaHecke(int d, int pos)
        {
            RemoveModDuplicates(d); //Makes mod d calculations
            //RemoveDuplicates(); //Removes possible zero exponents

            int m = Braiding[pos][1]; //The power to which the specified element is raised
            int index = Braiding[pos][0]; //The index of the element to be substituted from the quadratic relation
            List<Word> output = new List<Word>();

            Tuple<List<int[]>, List<int[]>> split = SplitYokonuma(pos);
            List<int[]> frame = new List<int[]>();
            Polynomial poly = new Polynomial();

            List<List<int[]>> words = new List<List<int[]>>();

            List<int[]> w;  w = new List<int[]>();

            if (m > 0)
            {
                if (m % 2 == 0)
                {
                    w = new List<int[]>();
                    for (int i = 0; i < split.Item1.Count; i++)
                        w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                    for (int i = 0; i < split.Item2.Count; i++)
                        w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                    output.Add(new Word(w, new Polynomial(1)));

                    for (int l = 0; l < m / 2; l++)
                    {
                        //poly.Add(new MultiVarTerm(1, new Tuple<char, int>('u', 2 * l + 1), new Tuple<char, int>('d', -1)));
                        //poly.Add(new MultiVarTerm(-1, new Tuple<char, int>('u', 2 * l), new Tuple<char, int>('d', -1)));
                        poly.Add(new MultiVarTerm(1.0 / d, new Tuple<char, int>('u', 2 * l + 1)));
                        poly.Add(new MultiVarTerm(-1.0 / d, new Tuple<char, int>('u', 2 * l)));
                    }
                    for (int n = 0; n < d; n++)
                    {
                        w = new List<int[]>();
                        List<int[]> w2 = new List<int[]>();
                        for (int i = 0; i < split.Item1.Count; i++)
                        {
                            w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                            w2.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                        }
                        w.Add(new int[] { 't', index, n });
                        w2.Add(new int[] { 't', index, n });
                        w.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 'g', index, 1 });
                        for (int i = 0; i < split.Item2.Count; i++)
                        {
                            w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                            w2.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                        }
                        output.Add(new Word(w, poly));
                        output.Add(new Word(w2, -poly));
                        //frame.Clear();
                        //frame.Add(new int[] { index, n }); //t_i^m
                        //frame.Add(new int[] { index + 1, -n }); //t_i^-m
                        //output.Add(split[0] * new Word(frame, new List<int[]>(), poly) * split[1]);
                        //output.Add(split[0] * new Word(frame, new List<int[]>(), -poly) * new int[] { 'g', index, 1 } * split[1]);
                    }
                }
                else
                {
                    w = new List<int[]>();
                    for (int i = 0; i < split.Item1.Count; i++)
                        w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                    w.Add(new int[] {'g', index, 1});
                    for (int i = 0; i < split.Item2.Count; i++)
                        w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                    output.Add(new Word(w, new Polynomial(1)));

                    for (int l = 0; l < m / 2; l++)
                    {
                        //poly.Add(new MultiVarTerm(1, new Tuple<char, int>('u', 2 * l + 2), new Tuple<char, int>('d', -1)));
                        //poly.Add(new MultiVarTerm(-1, new Tuple<char, int>('u', 2 * l + 1), new Tuple<char, int>('d', -1)));
                        poly.Add(new MultiVarTerm(1.0 / d, new Tuple<char, int>('u', 2 * l + 2)));
                        poly.Add(new MultiVarTerm(-1.0 / d, new Tuple<char, int>('u', 2 * l + 1)));
                    }
                    for (int n = 0; n < d; n++)
                    {
                        w = new List<int[]>();
                        List<int[]> w2 = new List<int[]>();
                        for (int i = 0; i < split.Item1.Count; i++)
                        {
                            w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                            w2.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                        }
                        w.Add(new int[] { 't', index, n });
                        w2.Add(new int[] { 't', index, n });
                        w.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 'g', index, 1 });
                        for (int i = 0; i < split.Item2.Count; i++)
                        {
                            w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                            w2.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                        }
                        output.Add(new Word(w, -poly));
                        output.Add(new Word(w2, poly));

                        //frame.Clear();
                        //frame.Add(new int[] { index, n }); //t_i^m
                        //frame.Add(new int[] { index + 1, -n }); //t_i^-m

                        //output.Add(split[0] * new Word(frame, new List<int[]>(), -poly) * split[1]);
                        //output.Add(split[0] * new Word(frame, new List<int[]>(), poly) * new int[] { 'g', index, 1 } * split[1]);
                    }
                }
            }
            else //Cannot be m=0 because zero exponents are removed from RemoveDuplicates()
            {
                if (m % 2 == 0)
                {
                    w = new List<int[]>();
                    for (int i = 0; i < split.Item1.Count; i++)
                        w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                    for (int i = 0; i < split.Item2.Count; i++)
                        w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                    output.Add(new Word(w, new Polynomial(1)));
                    //output.Add(split[0] * split[1]);

                    for (int l = 0; l < Math.Abs(m / 2); l++)
                    {
                        //poly.Add(new MultiVarTerm(1, new Tuple<char, int>('u', -2 * l - 2), new Tuple<char, int>('d', -1)));
                        //poly.Add(new MultiVarTerm(-1, new Tuple<char, int>('u', -2 * l - 1), new Tuple<char, int>('d', -1)));
                        poly.Add(new MultiVarTerm(1.0 / d, new Tuple<char, int>('u', -2 * l - 2)));
                        poly.Add(new MultiVarTerm(-1.0 / d, new Tuple<char, int>('u', -2 * l - 1))); //1/d [u^(2l+1) - u^(2l)]
                    }
                    for (int n = 0; n < d; n++)
                    {
                        w = new List<int[]>();
                        List<int[]> w2 = new List<int[]>();
                        for (int i = 0; i < split.Item1.Count; i++)
                        {
                            w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                            w2.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                        }
                        w.Add(new int[] { 't', index, n });
                        w2.Add(new int[] { 't', index, n });
                        w.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 'g', index, 1 });
                        for (int i = 0; i < split.Item2.Count; i++)
                        {
                            w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                            w2.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                        }
                        output.Add(new Word(w, poly));
                        output.Add(new Word(w2, -poly));
                        //frame.Clear();
                        //frame.Add(new int[] { index, n }); //t_i^m
                        //frame.Add(new int[] { index + 1, -n }); //t_i^-m

                        //output.Add(split[0] * new Word(frame, new List<int[]>(), poly) * split[1]);
                        //output.Add(split[0] * new Word(frame, new List<int[]>(), -poly) * new int[] { 'g', index, 1 } * split[1]);
                    }
                }
                else
                {
                    w = new List<int[]>();
                    for (int i = 0; i < split.Item1.Count; i++)
                        w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                    w.Add(new int[] { 'g', index, 1 });
                    for (int i = 0; i < split.Item2.Count; i++)
                        w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                    output.Add(new Word(w, new Polynomial(1)));
                    //output.Add(split[0] * new int[] { 'g', index, 1 } * split[1]);

                    for (int l = 0; l < Math.Abs(m / 2) + 1; l++)
                    {
                        //poly.Add(new MultiVarTerm(1, new Tuple<char, int>('u', -2 * l - 1), new Tuple<char, int>('d', -1)));
                        //poly.Add(new MultiVarTerm(-1, new Tuple<char, int>('u', -2 * l), new Tuple<char, int>('d', -1)));
                        poly.Add(new MultiVarTerm(1.0 / d, new Tuple<char, int>('u', -2 * l - 1)));
                        poly.Add(new MultiVarTerm(-1.0 / d, new Tuple<char, int>('u', -2 * l))); //1/d [u^(2l+1) - u^(2l)]
                    }
                    for (int n = 0; n < d; n++)
                    {
                        w = new List<int[]>();
                        List<int[]> w2 = new List<int[]>();
                        for (int i = 0; i < split.Item1.Count; i++)
                        {
                            w.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                            w2.Add(new int[] { split.Item1[i][0], split.Item1[i][1], split.Item1[i][2] });
                        }
                        w.Add(new int[] { 't', index, n });
                        w2.Add(new int[] { 't', index, n });
                        w.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 't', index + 1, -n });
                        w2.Add(new int[] { 'g', index, 1 });
                        for (int i = 0; i < split.Item2.Count; i++)
                        {
                            w.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                            w2.Add(new int[] { split.Item2[i][0], split.Item2[i][1], split.Item2[i][2] });
                        }
                        output.Add(new Word(w, -poly));
                        output.Add(new Word(w2, poly));
                        //frame.Clear();
                        //frame.Add(new int[] { index, n }); //t_i^m
                        //frame.Add(new int[] { index + 1, -n }); //t_i^-m

                        //output.Add(split[0] * new Word(frame, new List<int[]>(), -poly) * split[1]);
                        //output.Add(split[0] * new Word(frame, new List<int[]>(), poly) * new int[] { 'g', index, 1 } * split[1]);
                    }
                }
            }

            foreach (Word wd in output)
            {
                wd.RemoveModDuplicates(d);
                wd.RemoveDuplicates();
            }
            return output;
        }

        /// <summary>
        /// Calculates the trace for the Yokonuma-Hecke algebra
        /// </summary>
        /// <param name="d">The d integer</param>
        /// <returns>A Polynomial object</returns>
        public Polynomial TraceYokonumaHecke(int d)
        {
            RemoveModDuplicates(d); //Modulo d exponents
            RemoveDuplicates();

            //for (int i = 1; i < Braiding.Count; i++)
            //{
            //    if (Braiding[i - 1][0] < Braiding[i][0] - 1)
            //    {
            //        int[] temp = Braiding[i - 1];
            //        Braiding[i - 1] = Braiding[i];
            //        Braiding[i] = temp;
            //        this.RemoveDuplicates();
            //    }
            //}

            int[] frame = FindMaxIndexes(Framing);

            int exp;
            if (frame.Length == 0) { exp = 0; } //No framing part
            else { exp = Framing[frame[0]][1]; }

            SearchCommute(FindMaxIndexes(Braiding),true);
            int[] braid = FindMaxIndexes(Braiding);

            if (exp == 0 && Braiding.Count == 0) //Simple case
                return Coeff * TraceYokonumaHeckeUnit(0, false);
            else if (exp == 0 && Braiding.Count == 1 && Braiding[0][1] == 1) //Simple case
                return Coeff * TraceYokonumaHeckeUnit(0, true); //Simple case
            else if (exp == 0 && Braiding.Count == 1 && Braiding[0][1] != 1) //Quadratic relation needed
            {
                return ExpOrBridge(d);
            }
            else if (exp == 0 && braid.Length == 1 && !HasPowers(Braiding)) //Simple case
            {
                Word newWord = new Word(Framing, Braiding, new Polynomial(1));
                newWord.Braiding.RemoveAt(braid[0]);
                return Coeff * TraceYokonumaHeckeUnit(0, true) * newWord.TraceYokonumaHecke(d);
            }
            else if (exp > 0)
            {
                int frameIndex = Framing[frame[0]][0];
                bool max = true;
                for (int i = 0; i < Braiding.Count; i++)
                    if (frameIndex <= Braiding[i][0] + 1)
                        max = false;

                if (max) //The framing element t_i^m comes out as x_m
                {
                    Word newWord = new Word(Framing, Braiding, new Polynomial(1));
                    newWord.Framing.RemoveAt(frame[0]);
                    return Coeff * TraceYokonumaHeckeUnit(exp, false) * newWord.TraceYokonumaHecke(d);
                }
                else if (braid.Length == 1 && frameIndex <= Braiding[braid[0]][0] && Braiding[braid[0]][1] == 1) //The braiding element g_i comes out as z
                {
                    Word newWord = new Word(Framing, Braiding, new Polynomial(1));
                    newWord.Braiding.RemoveAt(braid[0]);
                    return Coeff * TraceYokonumaHeckeUnit(0, true) * newWord.TraceYokonumaHecke(d);
                }
                else if (braid.Length == 1 && frameIndex == Braiding[braid[0]][0] + 1 && Braiding[braid[0]][1] == 1)
                {
                    //Console.WriteLine(ToString());
                    List<int[]> w = new List<int[]>();

                    int current = frameIndex - 1;

                    for (int i = braid[0] - 1; i > -1; i--)
                    {
                        if (Braiding[i][0] == current - 1 && Braiding[i][1] % 2 != 0)
                            current --;
                        else if (Braiding[i][0] == current && Braiding[i][1] % 2 != 0)
                            current ++;
                    }

                    //int count = 0;
                    //for (int i = 0; i < braid[0]; i++)
                    //    if (Braiding[i][0] == Braiding[braid[0]][0] - 1)
                    //        count += Braiding[i][1];

                    for (int i = 1; i < Framing.Count; i++)
                        w.Add(new int[] { 't', Framing[i][0], Framing[i][1] });
                    w.Add(new int[] { 't', current, Framing[frame[0]][1] }); //wrong
                    for (int i = 0; i < Braiding.Count; i++)
                        if (i != braid[0]) { w.Add(new int[] { 'g', Braiding[i][0], Braiding[i][1] }); }

                    Word newWord = new Word(w);
                    //Console.WriteLine(newWord.ToString());
                    //Console.WriteLine();
                    return Coeff * TraceYokonumaHeckeUnit(0, true) * newWord.TraceYokonumaHecke(d);

                }
                else
                    return ExpOrBridge(d);
            }
            else
                return ExpOrBridge(d);
        }

        private Polynomial ExpOrBridge(int d)
        {
            String str1 = this.ToString();
            SearchCommute(FindMaxIndexes(Braiding), true);
            String str2 = this.ToString();

            if (HasPowers(Braiding)) //Powers are present
            {
                int[] maxExp = FindPowers(Braiding);
                List<Word> words = QuadraticRelationYokonumaHecke(d, maxExp[0]);
                Polynomial poly = words[0].TraceYokonumaHecke(d);
      
                for (int j = 1; j < words.Count; j++)
                    poly += words[j].TraceYokonumaHecke(d);
                //Polynomial[] polys = new Polynomial[words.Count - 1];
                //Parallel.For(1, words.Count, i =>
                //{
                //    polys[i - 1] = words[i].TraceYokonumaHecke(d);
                //});
                //for (int j = 1; j < words.Count; j++)
                //    poly += polys[j - 1];
                return Coeff * poly;
            }
            else //No powers at braiding part
            {
                //Bridge check
                Tuple<int, int> index = FindBridge(); //Searches for a bridge
                int[] frame = FindMaxIndexes(Framing);

                int exp;
                if (frame.Length == 0) { exp = 0; } //No framing part
                else { exp = Framing[frame[0]][1]; }
                if (exp == 0 && Braiding.Count == 0) //Simple case
                    return Coeff * TraceYokonumaHeckeUnit(0, false);
                else if (exp == 0 && Braiding.Count == 1 && Braiding[0][1] == 1) //Simple case
                    return Coeff * TraceYokonumaHeckeUnit(0, true); //Simple case

                if (index.Item1 == -1) //MUST NOT BE REACHED: if reached, there is a set of braids whose trace can not be calculated
                {
                    bool change = false;

                    for (int i = 1; i < Braiding.Count; i++)
                    {
                        if (Braiding[i - 1][0] < Braiding[i][0] - 1)
                        {
                            int[] temp = Braiding[i - 1];
                            Braiding[i - 1] = Braiding[i];
                            Braiding[i] = temp;
                            this.RemoveDuplicates();
                            change = true;
                        }
                    }
                    if (change)
                    {
                        return TraceYokonumaHecke(d);
                    }
                    else
                        throw new Exception("The trace for this word cannot be calculated"); //Exception thrown
                }
                else //Bridge found
                {
                    ReplaceBridge(index.Item1, index.Item2); //Replaces the bridge with a mountain
                    Polynomial poly = TraceYokonumaHecke(d);
                    return poly; //Calls recursively the trace function
                }
            }
        }

        #endregion

        /// <summary>
        /// Creates a string representation of the Word
        /// </summary>
        /// <returns>The Word as a string</returns>
        public override string ToString()
        {
            String str = ""; //Empty string

            //Framing part
            for (int i = 0; i < Framing.Count; i++)
            {
                str += "t" + Framing[i][0];
                if (Framing[i][1] != 1) { str += "^" + Framing[i][1]; }
                str += " ";
            }

            //Braiding part
            for (int i = 0; i < Braiding.Count; i++)
            {
                str += "g" + Braiding[i][0];
                if (Braiding[i][1] != 1) { str += "^" + Braiding[i][1]; }
                str += " ";
            }

            if (str.Length > 0) { str = str.Remove(str.Length - 1); }

            return str;
        }

        public int Degree()
        {
            int max = 0;
            int sum = 0;
            for(int i = 0; i < Braiding.Count; i++)
            {
                if (Braiding[i][0] > max) { max = Braiding[i][0]; }
                sum += Braiding[i][1];
            }
            return sum - max - 1;
        }
    }
}
